import pathlib

import numpy as np

from morpho_symm.data.DynamicsRecording import DynamicsRecording

# HOW DID I SAVED THING??
# temp_state =  np.zeros((1, 25))
# temp_state[0, 0] = state_current["position"][0]   WORLD
# temp_state[0, 1] = state_current["position"][1]   WORLD
# temp_state[0, 2] = state_current["position"][2]   WORLD
# temp_state[0, 3:6] = state_current["linear_velocity"][0:3] WORLD
# temp_state[0, 6] = state_current["orientation"][0] WORLD
# temp_state[0, 7] = state_current["orientation"][1] WORLD
# temp_state[0, 8] = state_current["orientation"][2] WORLD
# temp_state[0, 9] = state_current["angular_velocity"][0] BASE
# temp_state[0, 10] = state_current["angular_velocity"][1] BASE
# temp_state[0, 11] = state_current["angular_velocity"][2] BASE
# temp_state[0, 12:15] = state_current["foot_FL"] WORLD
# temp_state[0, 15:18] = state_current["foot_FR"] WORLD
# temp_state[0, 18:21] = state_current["foot_RL"] WORLD
# temp_state[0, 21:24] = state_current["foot_RR"] WORLD
# temp_state[0, 24:24+12] = d.qpos[6:18]
# temp_state[0, 36:36+12] = d.qvel[6:18]
# temp_state[0, 48] = pgg.step_freq
# data_state.append(copy.deepcopy(temp_state))


# temp_ref = np.zeros((1, 22))
# temp_ref[0, 0] = reference_state["ref_position"][2]
# temp_ref[0, 1:4] = reference_state["ref_linear_velocity"][0:3] WORLD - from a base one that is always costant here
# (0.4, 0, 0), but then is rotated in world
# temp_ref[0, 4] = reference_state["ref_orientation"][0]
# temp_ref[0, 5] = reference_state["ref_orientation"][1]
# temp_ref[0, 6] = reference_state["ref_angular_velocity"][0]
# temp_ref[0, 7] = reference_state["ref_angular_velocity"][1]
# temp_ref[0, 8] = reference_state["ref_angular_velocity"][2]
# temp_ref[0, 9:12] = reference_state["ref_FL_foot"]
# temp_ref[0, 12:15] = reference_state["ref_FR_foot"]
# temp_ref[0, 15:18] = reference_state["ref_RL_foot"]
# temp_ref[0, 18:21] = reference_state["ref_RR_foot"]
# temp_ref[0, 21] = 1.3 #nominal step frequency
# data_reference.append(copy.deepcopy(temp_ref))


# temp_input = np.zeros((1, 24))
# temp_input[0, 0:3] = nmpc_GRFs[0:3] WORLD
# temp_input[0, 3:6] = nmpc_GRFs[3:6] WORLD
# temp_input[0, 6:9] = nmpc_GRFs[6:9] WORLD
# temp_input[0, 9:12] = nmpc_GRFs[9:12] WORLD
# temp_input[0, 12:15] = tau_FL
# temp_input[0, 15:18] = tau_FR
# temp_input[0, 18:21] = tau_RL
# temp_input[0, 21:24] = tau_RR
# data_input.append(copy.deepcopy(temp_input))


# temp_disturbance = np.zeros((1, 7))
# temp_disturbance[0, 0:6] = disturbance_wrench WORLD
# temp_disturbance[0, 6] = start_disturbance_boolean WORLD
# data_external_disturbance.append(copy.deepcopy(temp_disturbance))


def convert_mini_cheetah_sb_controller_recording(data_path: pathlib.Path):
    """Conversion script for the recordings of observations from the Mini-Cheetah Robot.

    This function takes recordings stored into multiple numpy arrays for state, reference, input, and external
    disturbance. Each numpy array has its specific shape and dimensions:

    1. State (numpy array of shape (time, 85)):
        - base position (3,) at indices [0:3], representation: rep_Rd
        - base linear velocity (3,) at indices [3:6], representation: rep_Rd
        - base orientation (3,) at indices [6:9], representation: rep_euler_xyz
        - base angular velocity (3,) at indices [9:12], representation: rep_euler_xyz
        - feet positions (12,) at indices [12:24], (FL, FR, RL, RR) representation: rep_Rd_on_limbs
        - joint positions (12,) at indices [24:36], representation: rep_TqQ_js
        - joint velocities (12,) at indices [36:48], representation: rep_TqQ_js
        - step frequency (1,) at index [48], representation: rep_trivial

    2. Reference (numpy array of shape (time, 22)):
        - reference base height (1,) at index [0], representation: rep_z
        - reference base linear velocity (3,) at indices [1:4], representation: rep_Rd
        - reference base orientation (2,) at indices [4:6], representation: rep_euler_xyz
        - reference base angular velocity (3,) at indices [6:9], representation: rep_euler_xyz
        - reference feet positions (12,) at indices [9:21], representation: rep_Rd_on_limbs
        - nominal step frequency (1,) at index [21], representation: UNSURE

    3. Input (numpy array of shape (time, 24)):
        - ground reaction forces (12,) at indices [0:12], (FL, FR, RL, RR) representation: rep_Rd_on_limbs
        - joint torques (12,) at indices [12:24], representation: rep_TqQ_js

    4. External Disturbance (numpy array of shape (time, 7)):
        - disturbance_force (3,) at indices [0:3], representation: rep_Rd
        - disturbance_torque (3,) at indices [3:6], representation: rep_Rd_pseudo
        - start disturbance boolean (1,) at index [6], representation: rep_trivial

    The conversion process takes these measurements and does the following:
        1. Stores them into a DynamicsRecording format for easy loading.
        2. Defines the group representation for each observation.
        3. Changes joint position to Pinocchio convention, used by MorphoSymm.
    """
    assert data_path.exists(), f"Path {data_path.absolute()} does not exist"

    with open(data_path, "rb") as f:
        state = np.squeeze(np.load(f))
        reference = np.squeeze(np.load(f))
        input = np.squeeze(np.load(f))
        external_disturbance = np.squeeze(np.load(f))  # (n_samples, 7) : 7: 6 for the wrench, 1 for the boolean

    # Ensure the data has the expected shapes:
    assert state.shape[1] == 85, f"State shape is {state.shape}, expected (time, 85)"
    assert reference.shape[1] == 22, f"Reference shape is {reference.shape}, expected (time, 22)"
    assert input.shape[1] == 24, f"Input shape is {input.shape}, expected (time, 24)"
    assert external_disturbance.shape[1] == 7, (
        f"External Disturbance shape is {external_disturbance.shape}, expected (time, 7)"
    )

    # Here you can add the transformations and checks on the data as needed

    # Define the dataset.
    data_recording = DynamicsRecording(
        description=f"Mini Cheetah {data_path.parent.parent.stem}",
        info=dict(num_traj=1, trajectory_length=state.shape[0]),
        dynamics_parameters=dict(dt=0.001 * dt_subsample, group=dict(group_name=G.name, group_order=G.order())),
        recordings=dict(state=state, reference=reference, input=input, external_disturbance=external_disturbance),
        state_obs=(
            "joint_pos",
            "joint_vel",
            "base_z_error",
            "base_ori",
            "base_ori_error",
            "base_vel_error",
            "base_ang_vel_error",
        ),
        action_obs=("joint_torques",),
        obs_representations=dict(),  # Add the representations for each observation here
        obs_moments=dict(),  # Add the moments for each observation here
    )

    # Compute the mean and variance of all observations considering symmetry constraints.
    for obs_name in data_recording.recordings.keys():
        if obs_name in data_recording.obs_moments:
            continue
        data_recording.compute_obs_moments(obs_name=obs_name)

    file_name = f"n_trajs={data_recording.info['num_traj']}-frames={data_recording.info['trajectory_length']}.pkl"
    data_recording.save_to_file(data_path.parent.parent / file_name)
    print(f"Dynamics Recording saved to {data_path.parent.parent / file_name}")
    return data_recording


if __name__ == "__main__":
    data_path = pathlib.Path("./data_estimator.npy")
    data_recording = convert_mini_cheetah_sb_controller_recording(data_path)
    print("done")

    # Read the saved data

with open("./data_estimator.npy", "rb") as f:
    state = np.squeeze(np.load(f))
    reference = np.squeeze(np.load(f))
    input = np.squeeze(np.load(f))
    external_disturbance = np.squeeze(np.load(f))  # (n_samples, 7) : 7: 6 for the wrench, 1 for the boolean

    # Create a scatterplot matrix of disturbances
    import matplotlib.pyplot as plt
    import pandas as pd
    import seaborn as sns

    not_disturbed = np.alltrue(np.isclose(external_disturbance[:, :6], 0, atol=1e-2, rtol=1e-2), axis=1)
    external_disturbance[:, -1] = np.logical_not(not_disturbed)
    df = pd.DataFrame(external_disturbance, columns=["Fx", "Fy", "Fz", "Mx", "My", "Mz", "is_disturbed"])
    sns.pairplot(df, hue="is_disturbed")
    plt.show()

    print(state.shape)



import logging
from typing import Union

import escnn
import numpy as np
import torch
from escnn.nn import EquivariantModule, FieldType, GeometricTensor

from morpho_symm.nn.EquivariantModules import IsotypicBasis
from morpho_symm.utils.robot_utils import load_symmetric_system

log = logging.getLogger(__name__)


class EMLP(EquivariantModule):
    """Equivariant Multi-Layer Perceptron (EMLP) model."""

    def __init__(
        self,
        in_type: FieldType,
        out_type: FieldType,
        num_hidden_units: int = 64,
        num_layers: int = 3,
        bias: bool = True,
        activation: Union[str, EquivariantModule] = "ELU",
        head_with_activation: bool = False,
        batch_norm: bool = False,
        batch_norm_kwargs: dict = dict(affine=False, track_running_stats=True),
    ):
        """Constructor of an Equivariant Multi-Layer Perceptron (EMLP) model.

        This utility class allows to easily instanciate a G-equivariant MLP architecture. As a convention, we assume
        every internal layer is a map between the input space: X and an output space: Y, denoted as
        z = σ(y) = σ(W x + b). Where x ∈ X, W: X -> Y, b ∈ Y. The group representation used for intermediate layer
        embeddings ρ_Y: G -> GL(Y) is defined as a sum of multiple regular representations:
        ρ_Y := ρ_reg ⊕ ρ_reg ⊕ ... ⊕ ρ_reg. Therefore, the number of `hidden layer's neurons` will be a multiple of |G|.
        Being the multiplicities of the regular representation: ceil(num_hidden_units/|G|)

        Args:
        ----
            in_type (escnn.nn.FieldType): Input field type containing the representation of the input space.
            out_type (escnn.nn.FieldType): Output field type containing the representation of the output space. If the
            output type representation is composed only of multiples of the trivial representation then the this model
            will default to a G-invariant function. This G-invariant function is composed of a G-equivariant feature
            extractor, a G-invariant pooling layer, extracting invariant features from the equivariant features, and a
            last linear (unconstrained) layer to map the invariant features to the output space.
                TODO: Create a G-invariant EMLP class where we can control the network processing the G-invariant
                 features instead of defaulting to a single linear layer.
            num_hidden_units: Number of hidden units in the intermediate layers. The effective number of hidden units
            will be ceil(num_hidden_units/|G|). Since we assume intermediate embeddings are regular fields.
            activation (escnn.nn.EquivariantModule, str): Name of pointwise activation function to use.
            num_layers: Number of layers in the MLP including input and output/head layers. That is, the number of
            hidden layers will be num_layers - 2.
            bias: Whether to include a bias term in the linear layers.
            activation (escnn.nn.EquivariantModule, list(escnn.nn.EquivariantModule)): If a single activation module is
            provided it will be used for all layers except the output layer. If a list of activation modules is provided
            then `num_layers` activation equivariant modules should be provided.
            head_with_activation: Whether to include an activation module in the output layer.
            init_mode: Not used until now. Will be used to initialize the weights of the MLP.
        """
        super(EMLP, self).__init__()
        logging.info("Instantiating EMLP (PyTorch)")
        self.in_type, self.out_type = in_type, out_type
        self.gspace = self.in_type.gspace
        self.group = self.gspace.fibergroup
        self.num_layers = num_layers

        if batch_norm:
            log.warning("Equivariant Batch norm affects the performance of the model. Dont use if for now!!!")
        # Check if the network is a G-invariant function (i.e., out rep is composed only of the trivial representation)
        out_irreps = set(out_type.representation.irreps)
        if len(out_irreps) == 1 and self.group.trivial_representation.id == list(out_irreps)[0]:
            self.invariant_fn = True
        else:
            self.invariant_fn = False
            input_irreps = set(in_type.representation.irreps)
            inner_irreps = set(out_type.irreps)
            diff = input_irreps.symmetric_difference(inner_irreps)
            if len(diff) > 0:
                log.warning(
                    f"Irreps {list(diff)} of group {self.gspace.fibergroup} are not in the input/output types."
                    f"This represents an information bottleneck. Consider extracting invariant features."
                )

        if self.num_layers == 1 and not head_with_activation:
            log.warning(f"{self} model with 1 layer and no activation. This is equivalent to a linear map")

        if isinstance(activation, str):
            # Approximate the num of neurons as the num of signals in the space spawned by the irreps of the input type
            # To compute the signal over the group we use all elements for finite groups
            hidden_activation = self.get_activation(activation, in_type=in_type, desired_hidden_units=num_hidden_units)
            hidden_type = hidden_activation.in_type
        elif isinstance(activation, EquivariantModule):
            hidden_type = activation.in_type
        else:
            raise ValueError(f"Activation type {type(activation)} not supported.")

        layer_in_type = in_type
        self.net = escnn.nn.SequentialModule()
        for n in range(self.num_layers - 1):
            layer_out_type = hidden_type

            block = escnn.nn.SequentialModule()
            block.add_module(
                f"linear_{n}: in={layer_in_type.size}-out={layer_out_type.size}",
                escnn.nn.Linear(layer_in_type, layer_out_type, bias=bias),
            )
            if batch_norm:
                (block.add_module(f"batchnorm_{n}", escnn.nn.IIDBatchNorm1d(layer_out_type, **batch_norm_kwargs)),)
            block.add_module(f"act_{n}", hidden_activation)

            self.net.add_module(f"block_{n}", block)
            layer_in_type = layer_out_type

        # Add final layer
        self.net_head = None
        if self.invariant_fn:
            self.net_head = torch.nn.Sequential()
            # TODO: Make the G-invariant pooling with Isotypic Basis a stand alone module.
            # Module describing the change of basis to an Isotypic Basis required for efficient G-invariant pooling
            self.change2isotypic_basis = IsotypicBasis(hidden_type)
            # Number of G-invariant features from net output equals the number of G-stable subspaces.
            num_inv_features = len(hidden_type.irreps)
            self.net_head.add_module(
                f"linear_{num_layers - 1}", torch.nn.Linear(num_inv_features, out_type.size, bias=bias)
            )
            if head_with_activation:
                if batch_norm:
                    (
                        self.net_head.add_module(
                            f"batchnorm_{num_layers - 1}", torch.nn.BatchNorm1d(out_type.size, **batch_norm_kwargs)
                        ),
                    )
                self.net_head.add_module(f"act_{num_layers - 1}", activation)
        else:  # Equivariant Network
            self.net_head = escnn.nn.SequentialModule()
            self.net_head.add_module(f"linear_{num_layers - 1}", escnn.nn.Linear(layer_in_type, out_type, bias=bias))
            if head_with_activation:
                if batch_norm:
                    (
                        self.net_head.add_module(
                            f"batchnorm_{num_layers - 1}", escnn.nn.IIDBatchNorm1d(out_type, **batch_norm_kwargs)
                        ),
                    )
                self.net_head.add_module(f"act_{num_layers - 1}", escnn.nn.ELU(in_type=out_type))
        # Test the entire model is equivariant.
        # self.net.check_equivariance()

    def forward(self, x: GeometricTensor) -> GeometricTensor:
        """Forward pass of the EMLP model."""
        equivariant_features = self.net(x)
        if self.invariant_fn:
            iso_equivariant_features = self.change2isotypic_basis(equivariant_features)
            invariant_features = self.irrep_norm_pooling(iso_equivariant_features.tensor, iso_equivariant_features.type)
            output = self.net_head(invariant_features)
            output = self.out_type(output)  # Wrap over invariant field type
        else:
            output = self.net_head(equivariant_features)
        return output

    def reset_parameters(self, init_mode=None):
        """Initialize weights and biases of E-MLP model."""
        raise NotImplementedError()

    @staticmethod
    def get_activation(activation, in_type: FieldType, desired_hidden_units: int) -> EquivariantModule:
        gspace = in_type.gspace
        group = gspace.fibergroup

        grid_kwargs = EMLP.get_group_kwargs(group)

        unique_irreps = set(in_type.irreps)
        unique_irreps_dim = sum([group.irrep(*id).size for id in set(in_type.irreps)])
        channels = int(np.ceil(desired_hidden_units // unique_irreps_dim))
        if "identity" in activation.lower():
            raise NotImplementedError("Identity activation not implemented yet")
            # return escnn.nn.IdentityModule()
        else:
            act = escnn.nn.FourierPointwise(
                gspace,
                channels=channels,
                irreps=list(unique_irreps),
                function=f"p_{activation.lower()}",
                inplace=True,
                **grid_kwargs,
            )
        # assert (act.out_type.size - desired_hidden_units) <= unique_irreps_dim, \
        #     f"out_type.size {act.out_type.size} - des_hidden_units {desired_hidden_units} > {unique_irreps_dim}"
        return act

    @staticmethod
    def get_group_kwargs(group: escnn.group.Group):
        grid_type = "regular" if not group.continuous else "rand"
        N = group.order() if not group.continuous else 10
        kwargs = dict()

        if isinstance(group, escnn.group.DihedralGroup):
            N = N // 2
        elif isinstance(group, escnn.group.DirectProductGroup):
            G1_args = EMLP.get_group_kwargs(group.G1)
            G2_args = EMLP.get_group_kwargs(group.G2)
            kwargs.update({f"G1_{k}": v for k, v in G1_args.items()})
            kwargs.update({f"G2_{k}": v for k, v in G2_args.items()})

        return dict(N=N, type=grid_type, **kwargs)

    @staticmethod
    def irrep_norm_pooling(x: torch.Tensor, field_type: FieldType) -> torch.Tensor:
        from morpho_symm.utils.rep_theory_utils import irreps_stats

        n_inv_features = len(field_type.irreps)
        # TODO: Ensure isotypic basis i.e irreps of the same type are consecutive to each other.
        inv_features = []
        for field_start, field_end, rep in zip(
            field_type.fields_start, field_type.fields_end, field_type.representations
        ):
            # Each field here represents a representation of an Isotypic Subspace. This rep is only composed of a single
            # irrep type.
            x_field = x[..., field_start:field_end]
            num_G_stable_spaces = len(rep.irreps)  # Number of G-invariant features = multiplicity of irrep
            # Again this assumes we are already in an Isotypic basis
            unique_irreps, _, _ = irreps_stats(rep.irreps)
            assert len(unique_irreps) >= 1, f"Field type is not an Isotypic Subspace irreps:{unique_irreps}"
            # This basis is useful because we can apply the norm in a vectorized way
            # Reshape features to [batch, num_G_stable_spaces, num_features_per_G_stable_space]
            x_field_p = torch.reshape(x_field, (x_field.shape[0], num_G_stable_spaces, -1))
            # Compute G-invariant measures as the norm of the features in each G-stable space
            inv_field_features = torch.norm(x_field_p, dim=-1)
            # Append to the list of inv features
            inv_features.append(inv_field_features)
        # Concatenate all the invariant features
        inv_features = torch.cat(inv_features, dim=-1)
        assert inv_features.shape[-1] == n_inv_features, f"Expected {n_inv_features} got {inv_features.shape[-1]}"
        return inv_features

    def evaluate_output_shape(self, input_shape: tuple[int, ...]) -> tuple[int, ...]:
        """Returns the output shape of the model given an input shape."""
        batch_size = input_shape[0]
        return batch_size, self.out_type.size


if __name__ == "__main__":
    # Load robot instance and its symmetry group
    robot_name = "mini_cheetah-k4"  # or any of the robots in the library (see `/morpho_symm/cfg/robot`)
    robot, G = load_symmetric_system(robot_name=robot_name)
    # We use ESCNN to handle the group/representation-theoretic concepts and for the construction of equivariant neural networks.
    gspace = escnn.gspaces.no_base_space(G)
    # Get the relevant group representations.
    rep_QJ = G.representations["Q_js"]  # Used to transform joint-space position coordinates q_js ∈ Q_js
    rep_TqQJ = G.representations["TqQ_js"]  # Used to transform joint-space velocity coordinates v_js ∈ TqQ_js
    rep_R3 = G.representations["Rd"]  # Used to transform the linear momentum l ∈ R3
    rep_R3_pseudo = G.representations["Rd_pseudo"]  # Used to transform the angular momentum k ∈ R3

    nom = torch.tensor([1, 0.7, -1.4, 1, 0.7, -1.4, 1, 0.7, -1.4, 1, 0.7, -1.4])
    print(nom)
    nom = torch.stack([torch.cos(nom), torch.sin(nom)], dim=-1).view(-1)
    for g in G.elements[1:]:
        sym = torch.from_numpy(rep_QJ(g)).float() @ nom.float()
    sym = sym.view(-1, 2)
    sym = torch.atan2(sym[..., 1], sym[..., 0])
    print(sym)

    nom = torch.tensor([1, 0.7, -1.4, 1, 0.7, -1.4, 1, 0.7, -1.4, 1, 0.7, -1.4])
    # Define the input and output FieldTypes using the representations of each geometric object.
    # in_type = escnn.nn.FieldType(gspace, [G.regular_representation] * 5)
    # Define the input and output FieldTypes using the representations of each geometric object.
    # Representation of x := [q, v] ∈ Q_js x TqQ_js => ρ_X_js(g) := ρ_Q_js(g) ⊕ ρ_TqQ_js(g) | g ∈ G
    in_type = FieldType(gspace, [rep_R3, rep_R3_pseudo, rep_R3, rep_R3, rep_R3_pseudo, rep_TqQJ, rep_TqQJ, rep_TqQJ])
    out_type = escnn.nn.FieldType(gspace, [G.trivial_representation] * 1)
    # Test Invariant EMLP
    emlp = EMLP(in_type, out_type, num_hidden_units=128, num_layers=3, activation="ReLU", head_with_activation=False)
    emlp.eval()  # Shut down batch norm
    x = in_type(torch.randn(1, in_type.size))
    y = emlp(x)
    import numpy as np

    # G = escnn.group.DihedralGroup(6)
    # gspace = escnn.gspaces.no_base_space(G)
    # # Test Invariant EMLP
    # in_type = escnn.nn.FieldType(gspace, [G.regular_representation] * 5)
    # out_type = escnn.nn.FieldType(gspace, [G.trivial_representation] * 6)
    # emlp = EMLP(in_type, out_type,
    #             num_hidden_units=128,
    #             num_layers=3,
    #             activation="ReLU",
    #             head_with_activation=False)
    # emlp.eval()  # Shut down batch norm
    # x = in_type(torch.randn(1, in_type.size))
    # y = emlp(x)
    #
    # for g in G.elements:
    #     g_x = in_type(in_type.transform_fibers(x.tensor, g))  # Compute g · x
    #     g_y = emlp(g_x)  # Compute g · y
    #     assert torch.allclose(y.tensor, g_y.tensor, rtol=1e-4, atol=1e-4), \
    #         f"{g} invariance failed {y.tensor} != {g_y.tensor}"
    #
    # # Test Equivariant EMLP
    # in_type = escnn.nn.FieldType(gspace, [G.regular_representation] * 5)
    # out_type = escnn.nn.FieldType(gspace, [G.regular_representation] * 2)
    # emlp = EMLP(in_type, out_type,
    #             num_hidden_units=128,
    #             num_layers=3,
    #             activation="ReLU",
    #             head_with_activation=False)
    # emlp.eval()  # Shut down batch norm
    #
    # x = in_type(torch.randn(1, in_type.size))
    # y = emlp(x)
    #
    # for g in G.elements:
    #     g_x = in_type(in_type.transform_fibers(x.tensor, g))  # Compute g · x
    #     g_y_gt = out_type(out_type.transform_fibers(y.tensor, g))  # Compute ground truth g · y
    #     g_y = emlp(g_x)  # Compute g · y
    #     assert torch.allclose(g_y_gt.tensor, g_y.tensor, rtol=1e-4, atol=1e-4), \
    #         f"{g} invariance failed {g_y_gt.tensor} != {g_y.tensor}"



import itertools
from typing import Callable, Dict, List, Union

import numpy as np
from escnn.group import CyclicGroup, DihedralGroup, DirectProductGroup, Group, GroupElement
from escnn.group.representation import Representation
from scipy.linalg import block_diag

from morpho_symm.utils.algebra_utils import permutation_matrix
from morpho_symm.utils.mysc import CallableDict


def generate_cyclic_rep(G: CyclicGroup, rep):
    """Generate cylic froup form a representation of its generator."""
    h = G.generators[0]
    # Check the given matrix representations comply with group axioms
    # assert not np.allclose(rep[h], rep[G.identity]), "Invalid generator: h=e"
    assert np.allclose(np.linalg.matrix_power(rep[h], G.order()), rep[G.identity]), (
        f"Invalid rotation generator h_ref^{G.order()} != I"
    )

    curr_g = h
    while len(rep) < G.order():  # Use generator to obtain all elements and element reps in group
        g = curr_g @ h
        rep[g] = rep[curr_g] @ rep[h]
        curr_g = g

    return rep


def generate_dihedral_rep(G: DihedralGroup, rep):
    """Generate dihedral group form a representation of its generators."""
    h_rot, h_ref = G.generators
    # Check the given matrix representations comply with group axioms
    # assert not np.allclose(rep[h_ref], rep[G.identity]), "Invalid reflection generator: h_ref=e"
    # assert not np.allclose(rep[h_rot], rep[G.identity]), "Invalid rotation generator: h_rot=e"
    assert np.allclose(rep[h_ref] @ rep[h_ref], rep[G.identity]), "Invalid reflection generator `h_ref @ h_ref != I`"
    assert np.allclose(np.linalg.matrix_power(rep[h_rot], G.order() // 2), rep[G.identity]), (
        f"Invalid rotation generator h_ref^{G.order} != I"
    )

    curr_g, curr_ref_g = h_rot, h_ref @ h_rot
    rep[curr_ref_g] = rep[h_ref] @ rep[h_rot]
    while len(rep) < G.order():  # Use generator to obtain all elements and element reps in group
        g = curr_g @ h_rot
        gr = curr_ref_g @ h_rot
        rep[g] = rep[curr_g] @ rep[h_rot]
        rep[gr] = rep[curr_ref_g] @ rep[h_rot]
        curr_g, curr_ref_g = g, gr

    return rep


def generate_direct_product_rep(G: DirectProductGroup, rep1, rep2):
    """Generate direct product group form the two representations of each group generators."""
    rep = {}
    for h1, h2 in itertools.product(rep1.keys(), rep2.keys()):
        g = G.pair_elements(h1, h2)
        rep[g] = rep1[h1] @ rep2[h2]
    return rep


def group_rep_from_gens(G: Group, rep_H: Dict[GroupElement, np.ndarray]) -> Representation:
    """Generate a representation of all group actions from the representations of the group generators H={h1, h2, ...}.

    Such that any group action `g` can be obtained by multiplying the generators `h1`, `h2`, ... in `H`.
    Being all `h` in `H`
    Args:
        G (Group): Group to generate representation for.
        rep_H (Dict[GroupElement, np.ndarray]): Representation of the group generators H={h1, h2, ...}.

    Returns:
        Representation: ESCNN Representation of group actions generated by the input rep..
    """
    if G.identity not in rep_H:
        rep_H[G.identity] = np.eye(list(rep_H.values())[0].shape[0])

    if isinstance(G, CyclicGroup):
        rep_H = generate_cyclic_rep(G, rep_H)
    elif isinstance(G, DihedralGroup):
        rep_H = generate_dihedral_rep(G, rep_H)
    elif isinstance(G, DirectProductGroup):
        # Extract the generators of first and second group, generate the groups independently and then combine them
        H1, H2 = zip(*[G.split_element(h) for h in G.generators])
        rep_G1 = {G.G1.identity: rep_H[G.pair_elements(G.G1.identity, G.G2.identity)]}
        rep_G2 = {G.G2.identity: rep_H[G.pair_elements(G.G1.identity, G.G2.identity)]}
        rep_G1.update({h1: rep_H[G.inclusion1(h1)] for h1 in H1})
        rep_G2.update({h2: rep_H[G.inclusion2(h2)] for h2 in H2})

        # generate each subgroup representation
        group_rep_from_gens(G.G1, rep_G1)
        group_rep_from_gens(G.G2, rep_G2)

        # Do direct product of the generated subgroups reps.
        rep_H = generate_direct_product_rep(G, rep_G1, rep_G2)
    else:
        raise NotImplementedError(f"Group {G} not implemented yet.")

    # Convert Dict[GroupElement, np.ndarray] to escnn `Representation`
    rep_escnn = escnn_representation_form_mapping(G, rep_H)
    return rep_escnn


def irreps_stats(irreps_ids):
    str_ids = [str(irrep_id) for irrep_id in irreps_ids]
    unique_str_ids, counts, indices = np.unique(str_ids, return_counts=True, return_index=True)
    unique_ids = [eval(s) for s in unique_str_ids]
    return unique_ids, counts, indices


def escnn_representation_form_mapping(
    G: Group,
    representation: Union[Dict[GroupElement, np.ndarray], Callable[[GroupElement], np.ndarray]],
    name: str = "reconstructed",
):
    """Get a ESCNN representation instance from a mapping from group elements to unitary matrices.

    Args:
        G (Group): Symmetry group of the representation.
        representation (Union[Dict[GroupElement, np.ndarray], Callable[[GroupElement], np.ndarray]]): Mapping from
            group elements to unitary matrices.
        name (str, optional): Name of the representation. Defaults to 'reconstructed'.

    Returns:
        representation (Representation): ESCNN representation instance.
    """
    if isinstance(representation, dict):
        rep = CallableDict(representation)
    else:
        rep = representation

    # Find Q such that `iso_cplx(g) = Q @ rep(g) @ Q^-1` is block diagonal with blocks being complex irreps.
    cplx_irreps, Q = cplx_isotypic_decomposition(G, rep)
    # Get the size and location of each cplx irrep in `iso_cplx(g)`
    cplx_irreps_size = [irrep(G.sample()).shape[0] for irrep in cplx_irreps]
    irrep_dim_start = np.cumsum([0] + cplx_irreps_size[:-1])
    # Compute the character table of the found complex irreps and of all complex irreps of G
    irreps_char_table = compute_character_table(G, cplx_irreps)

    # We need to identify which real ESCNN irreps are present in rep(g).
    # First, we decompose the Group's ESCNN real irreps into complex irreps.
    escnn_cplx_irreps_data = {}
    for re_irrep in G.irreps():
        # Find Q_sub s.t. `block_diag([cplx_irrep_i1(g), cplx_irrep_i2(g)...]) = Q @ re_irrep_i(g) @ Q^-1`
        irreps, Q_sub = cplx_isotypic_decomposition(G, re_irrep)
        char_table = compute_character_table(G, irreps)
        escnn_cplx_irreps_data[re_irrep] = dict(subreps=irreps, Q=Q_sub, char_table=char_table)

    # Then, we find which of the Group complex irreps are present in the input representation, and determine
    # each group complex irrep multiplicity. As the complex irreps forming a real irrep can be spread over the
    # dimensions of the input rep, we find a permutation matrix P such that all complex irreps associated with a real
    # irrep are contiguous in dimensions.trifinger
    oneline_perm, Q_isore2isoimg = [], []
    escnn_real_irreps = []
    for escnn_irrep, data in escnn_cplx_irreps_data.items():
        # Match complex irreps by their character tables.
        multiplicities, irrep_locs = map_character_tables(data["char_table"], irreps_char_table)
        subreps_start_dims = [irrep_dim_start[i] for i in irrep_locs]  # Identify start of blocks in `rep(g)`
        data.update(multiplicities=multiplicities, subrep_start_dims=subreps_start_dims)
        assert np.unique(multiplicities).size == 1, "Multiplicities error"
        multiplicity = multiplicities[0]
        for m in range(multiplicity):
            Q_isore2isoimg.append(data["Q"])  # Add transformation from Real irrep to complex irrep
            escnn_real_irreps.append(escnn_irrep)  # Add escnn irrep to the list for instanciation
            for subrep, rep_start_dims in zip(data["subreps"], subreps_start_dims):
                rep_size = subrep[G.sample()].shape[0] if isinstance(subrep, dict) else subrep.size
                oneline_perm += list(range(rep_start_dims[m], rep_start_dims[m] + rep_size))
    # As the complex irreps forming a real irrep can be spread over the dimensions of the input rep, we find a
    # permutation matrix P such that all complex irreps of a real irrep are contiguous in dimensions / in the same block
    P = permutation_matrix(oneline_notation=oneline_perm)
    # Then we use the known transformations `Q_sub` for each real irrep, to create a mapping from cplx to real irreps.
    # s.t. `iso_re(g) = Q_iso_cplx2iso_re @ block_diag([cplx_irrep_11(g),...,cplx_irrep_ij(g)]) @ Q_iso_cplx2iso_re^-1`
    Q_iso_cplx2iso_re = block_diag(*[Q_sub.conj().T for Q_sub in Q_isore2isoimg])

    # Assert the matrix `P` and `Q_iso_cplx2iso_re` turn complex irreps into real irreps
    # `iso_re(g) = (Q_iso_cplx2iso_re @ P) @ iso_cplx(g) @ (Q_iso_cplx2iso_re @ P)^-1`,
    for g in G.elements:
        iso_re_g = block_diag(*[irrep(g) for irrep in escnn_real_irreps])
        iso_cplx_g = block_diag(*[cplx_irrep(g) for cplx_irrep in cplx_irreps])
        rec_iso_re_g = (Q_iso_cplx2iso_re @ P) @ iso_cplx_g @ (Q_iso_cplx2iso_re @ P).conj().T
        error = np.abs(iso_re_g - rec_iso_re_g)
        assert np.isclose(error, 0).all(), "Error in the conversion of Real irreps to Complex irreps"

    # Now we have an orthogonal transformation between the input `rep` and `iso_re`.
    #                        |     iso_cplx(g)     |
    # (Q_iso_cplx2iso_re @ P @ Q) @ rep(g) @ (Q^-1 @ P^-1 @ Q_iso_cplx2iso_re^-1) = Q_re @ rep(g) @ Q_re^-1 = iso_re(g)
    Q_re = Q_iso_cplx2iso_re @ P @ Q

    assert np.allclose(Q_re @ Q_re.conj().T, np.eye(Q_re.shape[0])), "Q_re is not an orthogonal transformation"
    if np.allclose(np.imag(Q_re), 0):
        Q_re = np.real(Q_re)  # Remove numerical noise and ensure rep(g) is of dtype: float instead of cfloat

    # Then we have that `Q_re^-1 @ iso_re(g) @ Q_re = rep(g)`
    reconstructed_rep = Representation(
        G, name=name, irreps=[irrep.id for irrep in escnn_real_irreps], change_of_basis=Q_re.conj().T
    )

    # Test ESCNN reconstruction
    for g in G.elements:
        g_true, g_rec = rep(g), reconstructed_rep(g)
        error = np.abs(g_true - g_rec)
        error[error < 1e-10] = 0
        assert np.allclose(error, 0), f"Reconstructed rep do not match input rep. g={g}, error:\n{error}"
        assert np.allclose(np.imag(g_rec), 0), f"Reconstructed rep not real for g={g}: \n{g_rec}"

    return reconstructed_rep


def is_complex_irreducible(
    G: Group, representation: Union[Dict[GroupElement, np.ndarray], Callable[[GroupElement], np.ndarray]]
):
    """Check if a representation is complex irreducible.

    We check this by asserting weather non-scalar (no multiple of
    identity ) Hermitian matrix `H` exists, such that `H` commutes with all group elements' representation.
    If rho is irreducible, this function returns (True, H=I)  where I is the identity matrix.
    Otherwise, returns (False, H) where H is a non-scalar matrix that commutes with all elements' representation.
    """
    if isinstance(representation, dict):

        def rep(g):
            return representation[g]
    else:
        rep = representation

    # Compute the dimension of the representation
    n = rep(G.sample()).shape[0]

    # Run through all r,s = 1,2,...,n
    for r in range(n):
        for s in range(n):
            # Define H_rs
            H_rs = np.zeros((n, n), dtype=complex)
            if r == s:
                H_rs[r, s] = 1
            elif r > s:
                H_rs[r, s] = 1
                H_rs[s, r] = 1
            else:  # r < s
                H_rs[r, s] = 1j
                H_rs[s, r] = -1j

            # Compute H
            H = sum([rep(g).conj().T @ H_rs @ rep(g) for g in G.elements]) / G.order()

            # If H is not a scalar matrix, then it is a matrix that commutes with all group actions.
            if not np.allclose(H[0, 0] * np.eye(H.shape[0]), H):
                return False, H
    # No Hermitian matrix was found to commute with all group actions. This is an irreducible rep
    return True, np.eye(n)


def decompose_representation(
    G: Group, representation: Union[Dict[GroupElement, np.ndarray], Callable[[GroupElement], np.ndarray]]
):
    """Find the Hermitian matrix `Q` that block-diagonalizes the representation `rep` of group `G`.

    Such that `Q @ rep[g] @ Q^H = block_diag(rep_1[g], ..., rep_m[g])` for all `g` in `G`.
    """
    import networkx as nx
    from networkx import Graph

    eps = 1e-12
    if isinstance(representation, dict):

        def rep(g):
            return representation[g]
    else:
        rep = representation
    # Compute the dimension of the representation
    n = rep(G.sample()).shape[0]

    for g in G.elements:  # Ensure the representation is unitary/orthogonal
        error = np.abs((rep(g) @ rep(g).conj().T) - np.eye(n))
        assert np.allclose(error, 0), f"Rep {rep} is not unitary: rep(g)@rep(g)^H=\n{(rep(g) @ rep(g).conj().T)}"

    # Find Hermitian matrix non-scalar `H` that commutes with all group actions
    is_irred, H = is_complex_irreducible(G, rep)
    if is_irred:
        return [rep], np.eye(n)

    # Eigen-decomposition of matrix `H = P·A·P^-1` reveals the G-invariant subspaces/eigenspaces of the representations.
    eivals, eigvects = np.linalg.eigh(H, UPLO="L")
    P = eigvects.conj().T
    assert np.allclose(P.conj().T @ np.diag(eivals) @ P, H)

    # Eigendcomposition is not guaranteed to block_diagonalize the representation. An additional permutation of the
    # rows and columns od the representation might be needed to produce a Jordan block canonical form.
    # First: We want to identify the diagonal blocks. To find them we use the trick of thinking of the representation
    # as an adjacency matrix of a graph. The non-zero entries of the adjacency matrix are the edges of the graph.
    edges = set()
    decomposed_reps = {}
    for g in G.elements:
        diag_rep = P @ rep(g) @ P.conj().T  # Obtain block diagonal representation
        diag_rep[np.abs(diag_rep) < eps] = 0  # Remove rounding errors.
        non_zero_idx = np.nonzero(diag_rep)
        edges.update([(x_idx, y_idx) for x_idx, y_idx in zip(*non_zero_idx)])
        decomposed_reps[g] = diag_rep

    # Each connected component of the graph is equivalent to the rows and columns determining a block in the diagonal
    graph = Graph()
    graph.add_edges_from(set(edges))
    connected_components = [sorted(list(comp)) for comp in nx.connected_components(graph)]
    connected_components = sorted(connected_components, key=lambda x: (len(x), min(x)))  # Impose a canonical order
    # If connected components are not adjacent dimensions, say subrep_1_dims = [0,2] and subrep_2_dims = [1,3] then
    # We permute them to get a jordan block canonical form. I.e. subrep_1_dims = [0,1] and subrep_2_dims = [2,3].
    oneline_notation = list(itertools.chain.from_iterable([list(comp) for comp in connected_components]))
    PJ = permutation_matrix(oneline_notation=oneline_notation)
    # After permuting the dimensions, we can assume the components are ordered in dimension
    ordered_connected_components = []
    idx = 0
    for comp in connected_components:
        ordered_connected_components.append(tuple(range(idx, idx + len(comp))))
        idx += len(comp)
    connected_components = ordered_connected_components

    # The output of connected components is the set of nodes/row-indices of the rep.
    subreps = [CallableDict() for _ in connected_components]
    for g in G.elements:
        for comp_id, comp in enumerate(connected_components):
            block_start, block_end = comp[0], comp[-1] + 1
            # Transform the decomposed representation into the Jordan Cannonical Form (jcf)
            jcf_rep = PJ @ decomposed_reps[g] @ PJ.T
            # Check Jordan Cannonical Form TODO: Extract this to a utils. function
            above_block = jcf_rep[0:block_start, block_start:block_end]
            below_block = jcf_rep[block_end:, block_start:block_end]
            left_block = jcf_rep[block_start:block_end, 0:block_start]
            right_block = jcf_rep[block_start:block_end, block_end:]

            assert np.allclose(above_block, 0) or above_block.size == 0, "Non zero elements above block"
            assert np.allclose(below_block, 0) or below_block.size == 0, "Non zero elements below block"
            assert np.allclose(left_block, 0) or left_block.size == 0, "Non zero elements left of block"
            assert np.allclose(right_block, 0) or right_block.size == 0, "Non zero elements right of block"
            sub_g = jcf_rep[block_start:block_end, block_start:block_end]
            subreps[comp_id][g] = sub_g

    # Decomposition to Jordan Canonical form is accomplished by (PJ @ P) @ rep @ (PJ @ P)^-1
    Q = PJ @ P

    # Test decomposition.
    for g in G.elements:
        jcf_rep = block_diag(*[subrep[g] for subrep in subreps])
        error = np.abs(jcf_rep - (Q @ rep(g) @ Q.conj().T))
        assert np.allclose(error, 0), f"Q @ rep[g] @ Q^-1 != block_diag[{[f'rep{i},' for i in range(len(subreps))]}]"

    return subreps, Q


def compute_character_table(G: Group, reps: List[Union[Dict[GroupElement, np.ndarray], Representation]]):
    """Computes the character table of a group for a given set of representations."""
    n_reps = len(reps)
    table = np.zeros((n_reps, G.order()), dtype=complex)
    for i, rep in enumerate(reps):
        for j, g in enumerate(G.elements):
            table[i, j] = rep.character(g) if isinstance(rep, Representation) else np.trace(rep(g))
    return table


def map_character_tables(in_table: np.ndarray, reference_table: np.ndarray):
    """Find a representation of a group in the set of irreducible representations."""
    n_in_reps = in_table.shape[0]
    out_ids, multiplicities = [], []
    for in_id in range(n_in_reps):
        character_orbit = in_table[in_id, :]
        orbit_error = np.isclose(np.abs(reference_table - character_orbit), 0)
        match_idx = np.argwhere(np.all(orbit_error, axis=1)).flatten()
        multiplicity = len(match_idx)
        out_ids.append(match_idx), multiplicities.append(multiplicity)
    return multiplicities, out_ids


def cplx_isotypic_decomposition(G: Group, representation: Callable[[GroupElement], np.ndarray]):
    """Perform the isotypic decomposition of unitary representation, decomposing the rep into complex irreps.

    Args:
        G (Group): Symmetry group of the representation.
        representation (Union[Dict[GroupElement, np.ndarray], Callable[[GroupElement], np.ndarray]]): dict/mapping from
        group elements to matrices, or a function that takes a group element and returns a matrix.

    Returns:
        sorted_irreps (List[Dict[GroupElement, np.ndarray]]): List of complex irreducible representations, sorted in
        ascending order of dimension.
        Q (ndarray): Hermitian matrix such that Q @ rep[g] @ Q^-1 is block diagonal, with blocks `sorted_irreps`.:

    """
    if isinstance(representation, dict):

        def rep(g):
            return representation[g]
    else:
        rep = representation

    n = rep(G.sample()).shape[0]
    subreps, Q_internal = decompose_representation(G, rep)

    found_irreps = []
    Qs = []

    # Check if each subrepresentation can be further decomposed.
    for subrep in subreps:
        n_sub = subrep(G.sample()).shape[0]  # Dimension of sub representation
        is_irred, _ = is_complex_irreducible(G, subrep)
        if is_irred:
            found_irreps.append(subrep)
            Qs.append(np.eye(n_sub))
        else:
            # Find Q_sub such that Q_sub @ subrep[g] @ Q_sub^-1 is block diagonal, with blocks `sub_subrep`
            sub_subreps, Q_sub = cplx_isotypic_decomposition(G, subrep)
            found_irreps += sub_subreps
            Qs.append(Q_sub)

    # Sort irreps by dimension.
    P, sorted_irreps = sorted_jordan_cann_form(G, found_irreps)

    # If subreps were decomposable, then these get further decomposed with an additional Hermitian matrix such that:
    # Q @ rep[g] @ Q^-1 = block_diag[irreps] | Q = (Q_external @ Q_internal)
    Q_external = block_diag(*Qs)
    Q = P @ Q_external @ Q_internal

    # Test isotypic decomposition.
    assert np.allclose(Q @ Q.conj().T, np.eye(n)), "Q is not unitary."
    for g in G.elements:
        g_iso = block_diag(*[irrep[g] if isinstance(irrep, dict) else irrep(g) for irrep in sorted_irreps])
        error = np.abs(g_iso - (Q @ rep(g) @ Q.conj().T))
        assert np.allclose(error, 0), f"Q @ rep[g] @ Q^-1 != block_diag[irreps[g]], for g={g}. Error \n:{error}"

    return sorted_irreps, Q


def sorted_jordan_cann_form(G: Group, reps: List[Callable[[GroupElement], np.ndarray]]):
    """Sorts a list of representations in ascending order of dimension, and returns a permutation matrix P such that.

    Args:
        G (Group): Symmetry group of the representation.
        reps: List of representations to sort by dimension.

    Returns:
        P (ndarray): Permutation matrix sorting the input reps.
        reps (List[Union[Dict[GroupElement, np.ndarray], Representation]]): Sorted list of representations.
    """
    reps_idx = range(len(reps))
    reps_size = [rep(G.sample()).shape[0] for rep in reps]
    sort_order = sorted(reps_idx, key=lambda idx: reps_size[idx])
    if sort_order == list(reps_idx):
        return np.eye(sum(reps_size)), reps
    irrep_dim_start = np.cumsum([0] + reps_size[:-1])
    oneline_perm = []
    for idx in sort_order:
        rep_size = reps_size[idx]
        oneline_perm += list(range(irrep_dim_start[idx], irrep_dim_start[idx] + rep_size))
    P = permutation_matrix(oneline_perm)

    return P, [reps[idx] for idx in sort_order]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 3/5/22
# @Author  : Daniel Ordonez
# @email   : daniels.ordonez@gmail.com
import logging
from math import ceil

import numpy as np
import torch
import torch.nn as nn
from emlp.reps.representation import Rep
from scipy.sparse import block_diag, issparse

from morpho_symm.groups.SparseRepresentation import SparseRep
from morpho_symm.groups.SymmetryGroups import C2

from .EMLP import EquivariantBlock
from .EquivariantModules import BasisConv1d, BasisLinear, EquivariantModel

log = logging.getLogger(__name__)


class ContactECNN(EquivariantModel):
    def __init__(
        self,
        rep_in: Rep,
        rep_out: Rep,
        window_size=150,
        cache_dir=None,
        dropout=0.5,
        init_mode="fan_in",
        inv_dim_scale=0.0,
        bias=True,
    ):
        super(ContactECNN, self).__init__(rep_in, rep_out, cache_dir)
        self.rep_in = rep_in
        self.rep_out = rep_out
        self.hidden_G = rep_in.G
        self.init_mode = init_mode
        self.window_size = window_size
        self.dropout = dropout
        self.inv_dims_scale = inv_dim_scale

        self.in_invariant_dims = self.rep_in.G.n_inv_dims
        inv_in, inv_out = rep_in.G.n_inv_dims / rep_in.G.d, rep_out.G.n_inv_dims / rep_out.G.d
        n_intermediate_layers = 6
        inv_ratios = np.linspace(inv_in, inv_out, n_intermediate_layers + 2, endpoint=True) * self.inv_dims_scale

        # CNN reps
        rep_ch_64_1 = SparseRep(self.hidden_G.canonical_group(64, inv_dims=ceil(64 * inv_ratios[1])))
        rep_ch_64_2 = SparseRep(self.hidden_G.canonical_group(64, inv_dims=ceil(64 * inv_ratios[2])))
        rep_ch_128_1 = SparseRep(self.hidden_G.canonical_group(128, inv_dims=ceil(128 * inv_ratios[3])))
        rep_ch_128_2 = SparseRep(self.hidden_G.canonical_group(128, inv_dims=ceil(128 * inv_ratios[4])))

        # Group of the flatten feature vector, must comply with the 2D symmetry.
        block2_out_window = int(window_size / 4)
        G = C2(generators=block_diag([rep_ch_128_2.G.discrete_generators[0]] * block2_out_window))

        # MLP reps
        rep_in_mlp = SparseRep(G)
        rep_ch_2048 = SparseRep(self.hidden_G.canonical_group(2048, inv_dims=ceil(2048 * inv_ratios[5])))
        rep_ch_512 = SparseRep(self.hidden_G.canonical_group(512, inv_dims=ceil(512 * inv_ratios[6])))

        self.block1 = nn.Sequential(
            BasisConv1d(rep_in=self.rep_in, rep_out=rep_ch_64_1, kernel_size=3, stride=1, padding=1, bias=bias),
            nn.ReLU(),
            BasisConv1d(rep_in=rep_ch_64_1, rep_out=rep_ch_64_2, kernel_size=3, stride=1, padding=1, bias=bias),
            nn.ReLU(),
            nn.Dropout(p=self.dropout),
            nn.MaxPool1d(kernel_size=2, stride=2),
        )

        self.block2 = nn.Sequential(
            BasisConv1d(rep_in=rep_ch_64_2, rep_out=rep_ch_128_1, kernel_size=3, stride=1, padding=1, bias=bias),
            nn.ReLU(),
            BasisConv1d(rep_in=rep_ch_128_1, rep_out=rep_ch_128_2, kernel_size=3, stride=1, padding=1, bias=bias),
            nn.ReLU(),
            nn.Dropout(p=self.dropout),
            nn.MaxPool1d(kernel_size=2, stride=2),
        )

        self.fc = nn.Sequential(
            EquivariantBlock(rep_in=rep_in_mlp, rep_out=rep_ch_2048, activation=nn.ReLU),
            nn.Dropout(p=self.dropout),
            EquivariantBlock(rep_in=rep_ch_2048, rep_out=rep_ch_512, activation=nn.ReLU),
            nn.Dropout(p=self.dropout),
            BasisLinear(rep_in=rep_ch_512, rep_out=self.rep_out),
        )

        # Test Each block equivariance.
        # EquivariantModel.test_module_equivariance(self.block1, rep_in=self.rep_in, rep_out=rep_ch_64,
        #                                           in_shape=(1, 54, 150))
        # EquivariantModel.test_module_equivariance(self.block2, rep_in=rep_ch_64, rep_out=rep_ch_128,
        #                                           in_shape=(1, 64, 75))
        # EquivariantModel.test_module_equivariance(self.fc, rep_in=rep_in_mlp, rep_out=self.rep_out)

        self.reset_parameters(init_mode=init_mode)
        # Test entire model equivariance.
        self.test_module_equivariance(
            module=self, rep_in=self.rep_in, rep_out=self.rep_out, in_shape=(1, 150, rep_in.G.d)
        )
        self.save_cache_file()

    def forward(self, x):
        x = x.permute(0, 2, 1)
        block1_out = self.block1(x)
        block2_out = self.block2(block1_out)
        # Ensure flattening maintains symmetry constraints
        block2_out = block2_out.permute(0, 2, 1)
        block2_out_reshape = block2_out.reshape(block2_out.shape[0], -1)
        fc_out = self.fc(block2_out_reshape)
        return fc_out

    def get_hparams(self):
        return {
            "window_size": self.window_size,
            "rep_in": str(self.rep_in),
            "rep_out": str(self.rep_in),
            "hidden_group": str(self.hidden_G),
            "init_mode": self.init_mode,
            "inv_dims_scale": self.inv_dims_scale,
            "dropout": self.dropout,
        }

    def reset_parameters(self, init_mode=None, model=None):
        assert init_mode is not None or self.init_mode is not None
        self.init_mode = init_mode if init_mode is not None else self.init_mode

        x = self if model is None else model
        for module in x.children():
            if isinstance(module, torch.nn.Sequential):
                for m in module.children():
                    if isinstance(m, BasisConv1d):
                        m.reset_parameters(mode=self.init_mode, activation="ReLU")
                    elif isinstance(m, EquivariantBlock):
                        m.linear.reset_parameters(
                            mode=self.init_mode, activation=m.activation.__class__.__name__.lower()
                        )
                    elif isinstance(m, BasisLinear):
                        m.reset_parameters(mode=self.init_mode, activation="Linear")

        log.info(f"{self.model_class} initialized with mode: {self.init_mode}")

    @staticmethod
    def test_module_equivariance(module: torch.nn.Module, rep_in, rep_out, in_shape=None):
        module.eval()
        shape = (rep_in.G.d) if in_shape is None else in_shape
        x = torch.randn(shape)
        for g_in, g_out in zip(rep_in.G.discrete_generators, rep_out.G.discrete_generators):
            g_in, g_out = (g_in.todense(), g_out.todense()) if issparse(g_in) else (g_in, g_out)
            g_in = torch.tensor(np.asarray(g_in), dtype=torch.float32).unsqueeze(0)
            g_out = torch.tensor(np.asarray(g_out), dtype=torch.float32).unsqueeze(0)

            y = module.forward(x)

            if x.ndim == 3:
                xx = x.permute(0, 2, 1)
                g_x = (g_in @ xx.unsqueeze(1)).squeeze(1)
                g_x = g_x.permute(0, 2, 1)
            else:
                g_x = g_in @ x

            g_y_pred = module.forward(g_x)

            g_y_true = (g_out[0, :, :] @ y.T).T

            if not torch.allclose(g_y_true, g_y_pred, atol=1e-4, rtol=1e-4):
                error = g_y_true - g_y_pred
                raise RuntimeError(f"{module} is not equivariant to in/out group generators f(g·x) - g·y:{error}")
        module.train()

    def unfreeze_equivariance(self, num_layers=1):
        # Freeze most of model model.
        for parameter in self.parameters():
            parameter.requires_grad = False
        last_equiv_layers = self.fc[0::2]
        for e_layer in last_equiv_layers[-num_layers:]:
            e_layer.unfreeze_equivariance()



import logging
import pathlib
import time

import numpy as np
import torch
import torch.nn.functional as F
from escnn import gspaces
from escnn.group import Group, directsum
from escnn.nn import FieldType
from pinocchio import pinocchio_pywrap as pin
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate

from morpho_symm.data.DynamicsRecording import DynamicsRecording
from morpho_symm.robots.PinBulletWrapper import PinBulletWrapper

# from morpho_symm.utils.algebra_utils import dense
from morpho_symm.utils.robot_utils import load_symmetric_system

log = logging.getLogger(__name__)

np.set_printoptions(precision=4)


class ProprioceptiveDataset(Dataset):
    def __init__(
        self,
        robot_name,
        type="train",
        angular_momentum=True,
        augment=False,
        train_ratio=0.7,
        test_ratio=0.15,
        val_ratio=0.15,
        n_total_samples=10000,
        kinetic_energy=False,
        dtype=torch.float32,
        data_path="dataset/proprioceptive",
        device="cpu",
        debug=False,
    ):
        self.dataset_type = type
        self.dtype = dtype
        self.angular_momentum = angular_momentum
        self.kin_energy = kinetic_energy

        # Load robot, symmetry group and input-output field types/representations
        self.robot, self.G, self.in_type, self.out_type = self.define_input_output_field_types(robot_name)

        self._pb = None  # GUI debug
        self.augment = augment if isinstance(augment, bool) else False

        self._samples = n_total_samples
        self.dataset_path = pathlib.Path(data_path).joinpath(f"{self.robot.name}/n_samples_{n_total_samples:d}.plk")
        self.ensure_dataset_existance()
        partition_path = self.ensure_dataset_partition(train_ratio, test_ratio, val_ratio)

        # Load data
        assert type.lower() in ["train", "test", "val"], "type must be one of these [train, test, val]"
        file_path = partition_path.joinpath(f"{type}.npz")
        assert file_path.exists(), file_path.absolute()
        # Loading on multiple threads might lead to issues -------------------
        trials = 5
        data = None
        while data is None and trials > 0:
            try:
                data = np.load(str(file_path))
            except:
                trials -= 1
                time.sleep(np.random.random())
        #  -------------------------------------------------------------------
        if self.kin_energy:
            X, Y = data["X"], data["KinE"]
        else:
            X, Y = data["X"], data["Y"]

        q, dq = self.robot.get_init_config(random=False, fix_base=True)
        self.base_q = q[:7]
        self.base_dq = dq[:6]

        self.X = torch.from_numpy(X).type("torch.FloatTensor").to(device)
        self.Y = torch.from_numpy(Y).type("torch.FloatTensor").to(device)

        if debug:
            self.plot_statistics()

        # if isinstance(augment, str) and augment.lower() == "hard":
        #     for g in self.G.elements[1:]:
        #         rep_X = self.in_type.fiber_representation(g).to(self.X.device)
        #         rep_Y = self.out_type.fiber_representation(g).to(self.Y.device)
        #         gX = (rep_X @ self.X.T).T
        #         gY = (rep_Y @ self.Y.T).T
        #         self.X = torch.vstack([self.X, gX])
        #         self.Y = torch.vstack([self.Y, gY])

        self.loss_fn = F.mse_loss
        log.info(str(self))

        # TODO: Remove
        # self.test_equivariance()

    def test_equivariance(self):
        trials = 50
        for trial in range(trials):
            q, dq = self.robot.get_init_config(random=True, fix_base=True)
            dq = np.random.uniform(-np.pi, np.pi, size=dq.shape)
            kin_E = pin.computeKineticEnergy(self.robot.pinocchio_robot.model, self.robot.pinocchio_robot.data, q, dq)

            x = np.concatenate((q[7:], dq[6:]))
            x = x.astype(np.float64)
            y = self.get_hg(q[7:], dq[6:])
            y = y.astype(np.float64)

            x_orbit, y_orbit = [x], [y]
            y_true = [y]

            non_equivariance_detected = False
            rep_X = self.in_type.representation
            rep_Y = self.out_type.representation
            # Get all possible group actions
            for g in self.G.elements:
                gx, gy = rep_X(g) @ x, rep_Y(g) @ y
                x_orbit.append(gx)
                y_orbit.append(gy)
                assert gx.dtype == x.dtype, (gx.dtype, x.dtype)
                assert gy.dtype == y.dtype, (gy.dtype, y.dtype)

                qj, dqj = gx[: self.robot.nq - 7], gx[self.robot.nq - 7 :]
                gq, gdq = q, dq
                gq[7:] = qj
                gdq[6:] = dqj
                g_kinE = pin.computeKineticEnergy(
                    self.robot.pinocchio_robot.model, self.robot.pinocchio_robot.data, gq, gdq
                )

                # Ensure kinetic energy is preserved, up to 5% relative error
                kin_E_rel_error = np.abs((g_kinE - kin_E) / kin_E)
                if kin_E_rel_error > 0.05:
                    raise AttributeError(f"Kinetic energy G-invariance violated: relative error {g_kinE} != {kin_E}")

                gy_true = self.get_hg(*np.split(gx, 2))
                y_true.append(gy_true)

                assert gy_true.dtype == y.dtype, (gy_true.dtype, y.dtype)

                error = gy_true - gy
                rel_error_norm = np.linalg.norm(error) / np.linalg.norm(gy_true)
                cos_sim = np.dot(gy, gy_true) / (np.linalg.norm(gy_true) * np.linalg.norm(gy))

                # TODO: Change true
                if rel_error_norm > 0.05 and cos_sim <= 0.95:
                    non_equivariance_detected = True

            if non_equivariance_detected:
                try:
                    splited_orbits = [np.split(x, 2) for x in x_orbit]
                    qs = [x[0] for x in splited_orbits]
                    dqs = [x[1] for x in splited_orbits]
                    self.gui_debug(qs=qs, dqs=dqs, hgs=y_true, ghgs=y_orbit)
                except Exception as e:
                    raise e
                    logging.warning(f"Unable to start GUI of pybullet: {str(e)}")
                # raise AttributeError(f"Ground truth hg(q,dq) = Ag(q)dq is not equivariant to provided groups: \n" +
                #                      f"x:{x}\ng*x:{gx}\ny:{y} \ng*y:{gy}\n" +
                #                      f"Aq(g*q,g*dq):{gy_true}\nError:{error}")
        return None

    def compute_metrics(self, y, y_pred) -> dict:
        # Original scales and means.
        with torch.no_grad():
            metrics = {}

            assert y.shape == y_pred.shape, (y.shape, y_pred.shape)

            if self.kin_energy:
                metrics["kin_energy_err"] = torch.mean(torch.abs(y - y_pred))
                return metrics

            lin, lin_pred = y[:, :3], y_pred[:, :3]
            metrics["lin_cos_sim"] = torch.mean(F.cosine_similarity(lin, lin_pred, dim=-1))
            metrics["lin_err"] = torch.mean(torch.linalg.norm(lin - lin_pred, dim=-1))

            ang, ang_pred = y[:, 3:], y_pred[:, 3:]
            metrics["ang_cos_sim"] = torch.mean(F.cosine_similarity(ang, ang_pred, dim=-1))
            metrics["ang_err"] = torch.mean(torch.linalg.norm(ang - ang_pred, dim=-1))
        return metrics

    def __len__(self):
        return self.Y.shape[0]

    def __getitem__(self, i):
        if self.kin_energy:
            x, y = (
                self.X[i, :],
                self.Y[i],
            )
        else:
            if self.angular_momentum:
                x, y = (
                    self.X[i, :],
                    self.Y[i, :],
                )
            else:
                x, y = (
                    self.X[i, :],
                    self.Y[i, :3],
                )
        return x, y

    def collate_fn(self, batch):
        # Enforce data type in batched array
        # Small hack to do batched augmentation. TODO: Although efficient this should be done somewhere else.
        x_batch, y_batch = default_collate(batch)

        if self.augment:  # Sample uniformly among symmetry actions including identity
            g = self.G.sample()

            g_x_batch = self.in_type.transform_fibers(x_batch, g)
            g_y_batch = self.out_type.transform_fibers(y_batch, g)
            x_batch, y_batch = g_x_batch, g_y_batch
        return x_batch.to(self.dtype), y_batch.to(self.dtype)

    @property
    def augment(self):
        return self._augment

    @augment.setter
    def augment(self, augment: bool):
        self._augment = augment
        log.debug(f"Dataset Group Augmentation {'ACTIVATED' if self.augment else 'DEACTIVATED'}")

    def get_hg(self, qj, dqj):
        hg = self.robot.pinocchio_robot.centroidalMomentum(
            q=np.concatenate((self.base_q, qj)), v=np.concatenate((self.base_dq, dqj))
        )
        hg = np.array(hg)
        if self.angular_momentum:
            return hg
        return hg[:3]

    def ensure_dataset_existance(self):
        # if self.robot.
        q, dq = self.robot.get_init_config(random=False, fix_base=True)
        self.base_q = q[:7]
        self.base_dq = dq[:6]
        if self.dataset_path.exists():
            log.debug(f"Loading dataset of size {self._samples}")
        else:
            log.info(f"Generating dataset for {self.robot.__class__.__name__} of size {self._samples} samples")
            self.dataset_path.parent.mkdir(exist_ok=True, parents=True)
            # Ensure deterministic generation
            np.random.seed(29081995)
            # Get joint limits.
            dq_max = np.asarray(self.robot.velocity_limits)
            dq_max = np.minimum(dq_max, np.pi)
            q_min, q_max = self.robot.joint_pos_limits
            q_min = np.maximum(q_min, -np.pi)
            q_max = np.minimum(q_max, np.pi)
            print(f"vel max {dq_max}")
            print(f"qmin {q_min} \n qmax {q_max}")
            rep_Q = self.G.representations["Q_js"]
            rep_TqQ = self.G.representations["TqQ_js"]
            rep_R3 = self.G.representations["R3"]
            rep_R3_pseudo = self.G.representations["R3_pseudo"]
            # Rep for center of mass momentum y := [l, k] ∈ R3 x R3  =>    ρ_R3(g) ⊕ ρ_R3pseudo(g)  | g ∈ G
            com_rep = directsum([rep_R3, rep_R3_pseudo])

            x = np.zeros((self._samples, self.robot.nq + self.robot.nv - 7 - 6))
            y = np.zeros((self._samples, 6))
            kinE = np.zeros((self._samples, 1))
            for i in range(self._samples):
                q[7:] = np.random.uniform(q_min, q_max, size=None)
                dq[6:] = np.random.uniform(-dq_max, dq_max, size=None)
                hg = self.robot.pinocchio_robot.centroidalMomentum(q, dq)
                kin_energy = pin.computeKineticEnergy(
                    self.robot.pinocchio_robot.model, self.robot.pinocchio_robot.data, q, dq
                )
                kinE[i, :] = kin_energy
                y[i, :] = hg.np
                x[i, :] = np.concatenate((q[7:], dq[6:]))

            # Pinnochio introduces small but considerable equivariance numerical error, even when the robot kinematics
            # and dynamics are completely equivariant. So we make the gt the avg of the augmented predictions.
            ys_pin = [y]
            G_kinE = [kinE]
            for g in self.G.elements[1:]:
                gx = np.squeeze(self.in_type.representation(g) @ x.T).T
                # gy = np.squeeze(y @ g_out)
                gy_pin = np.zeros((self._samples, 6))
                g_kinE = np.zeros((self._samples, 1))
                # Generate random configuration samples.
                for i, x_sample in enumerate(gx):
                    qj, vj = x_sample[: self.robot.nq - 7], x_sample[self.robot.nq - 7 :]
                    q[7:] = qj
                    dq[6:] = vj
                    hg = self.robot.pinocchio_robot.centroidalMomentum(q, dq)
                    kin_energy = pin.computeKineticEnergy(
                        self.robot.pinocchio_robot.model, self.robot.pinocchio_robot.data, q, dq
                    )
                    gy_pin[i, :] = hg.np
                    g_kinE[i, :] = kin_energy

                # Inverse is not needed for the groups we use (C2, V4).
                ys_pin.append((com_rep(~g) @ gy_pin.T).T)
                G_kinE.append(g_kinE)

            y_pin_avg = np.mean(ys_pin, axis=0)
            y = y_pin_avg  # To mitigate numerical error
            mean_kinE = np.mean(G_kinE, axis=0)
            max_err = np.max(np.abs(kinE - mean_kinE))
            kinE = mean_kinE
            # From the augmented dataset take the desired samples.
            X, Y, KinE = x, y, kinE

            print(X.shape)
            print(np.expand_dims(X, 1).shape)
            data_record = DynamicsRecording(
                description=f"Proprioceptive data from {self.robot.name} robot",
                recordings={
                    "X": np.expand_dims(X, 1),
                    "Y": np.expand_dims(Y, 1),
                    "kinetic_energy": np.expand_dims(KinE, 1),
                    "qjs": np.expand_dims(X[..., : self.robot.nq - 7], 1),
                    "vjs": np.expand_dims(X[..., self.robot.nq - 7 :], 1),
                    "com_lin_momentum": np.expand_dims(Y[..., :3], 1),
                    "com_ang_momentum": np.expand_dims(Y[..., 3:], 1),
                },
                state_obs=("X"),
                obs_representations={
                    "X": rep_Q + rep_TqQ,
                    "Y": com_rep,
                    "kinetic_energy": self.G.trivial_representation,
                    "qjs": rep_Q,
                    "vjs": rep_TqQ,
                    "com_lin_momentum": rep_R3,
                    "com_ang_momentum": rep_R3_pseudo,
                },
            )
            data_record.save_to_file(self.dataset_path)
            print(f"Saved data record to {self.dataset_path}")
            # np.savez_compressed(str(self.dataset_path), X=X, Y=Y, KinE=KinE)
            log.info(f"Dataset saved to {self.dataset_path.absolute()}")
            self.test_equivariance()
        assert self.dataset_path.exists(), "Something went wrong"

    def ensure_dataset_partition(self, train_ratio=0.7, test_ratio=0.15, val_ratio=0.15) -> pathlib.Path:
        assert train_ratio + test_ratio + val_ratio <= 1.0
        partition_folder = f"{self._samples}_train={train_ratio:.2f}_test={test_ratio:.2f}_val={val_ratio:.2f}"
        partition_path = self.dataset_path.parent.joinpath(partition_folder)
        partition_path.mkdir(exist_ok=True)

        train_path, test_path, val_path = (
            partition_path.joinpath("train.npz"),
            partition_path.joinpath("test.npz"),
            partition_path.joinpath("val.npz"),
        )

        if not train_path.exists() or not val_path.exists() or not test_path.exists():
            data = np.load(str(self.dataset_path))
            X, Y, KinE = data["X"], data["Y"], data["KinE"]

            num_data = X.shape[0]
            num_train = int(train_ratio * num_data)
            num_val = int(val_ratio * num_data)
            num_test = int(test_ratio * num_data)
            assert train_ratio + val_ratio + test_ratio <= 1.0

            X_test, Y_test = X[:num_test, :], Y[:num_test, :]
            X_val, Y_val = X[num_test : num_val + num_test, :], Y[num_test : num_val + num_test, :]
            X_train, Y_train = X[-num_train:, :], Y[-num_train:, :]

            KinE_test = KinE[:num_test, :]
            KinE_val = KinE[num_test : num_val + num_test, :]
            KinE_train = KinE[-num_train:, :]

            # Use full group orbit on test set to account for equivariance/invariance error.
            G_X, G_Y, G_KinE = [X_test], [Y_test], [KinE_test]
            for g in self.G.elements[1:]:
                rep_X = self.in_type.representation
                rep_Y = self.out_type.representation
                gX = (rep_X(g) @ X_test.T).T
                gY = (rep_Y(g) @ Y_test.T).T
                G_X.append(gX)
                G_Y.append(gY)
                G_KinE.append(KinE_test)

            X_test = np.vstack(G_X)
            Y_test = np.vstack(G_Y)
            KinE_test = np.vstack(G_KinE)

            np.savez_compressed(str(partition_path.joinpath("train.npz")), X=X_train, Y=Y_train, KinE=KinE_train)
            np.savez_compressed(str(partition_path.joinpath("test.npz")), X=X_test, Y=Y_test, KinE=KinE_test)
            np.savez_compressed(str(partition_path.joinpath("val.npz")), X=X_val, Y=Y_val, KinE=KinE_val)

            log.info(f"Saving dataset partition {partition_folder} on {partition_path.parent}")
        else:
            log.debug(f"Loaded dataset partition {partition_folder}")
        return partition_path

    def to(self, device):
        log.info(f"Moving data to device:{device}")
        self.X.to(device)
        self.Y.to(device)

    def define_input_output_field_types(self, robot_name) -> (PinBulletWrapper, Group, FieldType, FieldType):
        """Define the input-output symmetry representations for the CoM function g·y = f(g·x) | g ∈ G.

        Define the symmetry representations for the Center of Mass momentum y := (l, k) ∈ R^3 x R^3, where l is the
        linear momentum and k is the angular momentum. The CoM momentum is a function of the system joint state
        x := (q_js, v_js) ∈ Q_js x TqQ_js, where Q_js is the joint space and TqQ_js is the tangent space of the
        joint space.

        Args:
            robot_cfg: The robot configuration (see cfg/robots).

        Returns:
            robot (PinBulletWrapper): The robot wrapper.
            gspace (GSpace): The total symmetry space of the system.
            input_type (FieldType): The input feature field type, describing the system state (q_js, v_js) and its
            symmetry transformations.
            output_type (FieldType): The output feature field type, describing the system CoM (hg) and its symmetry
            transformations.
        """
        robot, G = load_symmetric_system(robot_name=robot_name, debug=False)
        # For this application we compute the CoM w.r.t base frame, meaning that we ignore the fiber group Ed in which
        # the system evolves in:
        gspace = gspaces.no_base_space(G)
        # Get the relevant representations.
        rep_Q_js = G.representations["Q_js"]
        rep_TqQ_js = G.representations["TqQ_js"]
        rep_R3 = G.representations["R3"]
        rep_R3_pseudo = G.representations["R3_pseudo"]

        # Rep for x := [q, dq] ∈ Q_js x TqQ_js     =>    ρ_Q_js(g) ⊕ ρ_TqQ_js(g)  | g ∈ G
        in_type = FieldType(gspace, [rep_Q_js, rep_TqQ_js])

        if self.kin_energy:
            out_type = FieldType(gspace, [G.trivial_representation])
        else:  # CoM momentum
            # Rep for center of mass momentum y := [l, k] ∈ R3 x R3  =>    ρ_R3(g) ⊕ ρ_R3pseudo(g)  | g ∈ G
            out_type = FieldType(gspace, [rep_R3, rep_R3_pseudo])

        # TODO: handle subgroup cases.
        # if robot_cfg.gens_ids is not None and self.dataset_type in ["test", "val"]:
        #     Loaded symmetry group is not the full symmetry group of the robot
        #     Generate the true underlying data representations
        # robot_cfg = copy.copy(robot_cfg)
        # robot_cfg['gens_ids'] = None
        # _, rep_Ed_full, rep_QJ_full = load_robot_and_symmetries(robot_cfg)
        # rep_data_in = get_rep_data_in(rep_QJ_full)
        # rep_data_out = get_rep_data_out(rep_Ed_full)
        # log.info(f"[{self.dataset_type}] Dataset using the system's full symmetry group {type(rep_QJ_full.G)}")
        # return robot, rep_data_in, rep_data_out

        log.info(f"[{self.dataset_type}] Dataset using the symmetry group {type(G)}")
        return robot, G, in_type, out_type

    def __repr__(self):
        msg = (
            f"CoM Dataset: [{self.robot.robot_name}]-{self.dataset_type}-Aug:{self.augment}"
            f"-X:{self.X.shape}-Y:{self.Y.shape}"
        )
        return msg

    def __str__(self):
        return self.__repr__()

    def plot_statistics(self):
        import seaborn as sns
        from matplotlib import pyplot as plt

        x_orbits = [self.X.detach().cpu().numpy()]
        y_orbits = [self.Y.detach().cpu().numpy()]

        for g_in, g_out in self.t_group_actions[1:]:
            g_x = torch.matmul(self.X.unsqueeze(1), g_in.unsqueeze(0).to(self.X.dtype)).squeeze()
            g_y = torch.matmul(self.Y.unsqueeze(1), g_out.unsqueeze(0).to(self.Y.dtype)).squeeze()
            y_orbits.append(g_y.detach().cpu().numpy())
            x_orbits.append(g_x.detach().cpu().numpy())

        y_true = []
        for x_orbit in x_orbits:
            y_orbit = []
            for x in x_orbit:
                y = self.get_hg(*np.split(x, 2))
                y_orbit.append(y)
            y_true.append(np.vstack(y_orbit))

        i = 0
        errors = []
        for g_y, y_gt in zip(y_orbits, y_true):
            lin_error = np.linalg.norm(y_gt[:, :3] - g_y[:, :3], axis=1)
            ang_error = np.linalg.norm(y_gt[:, 3:] - g_y[:, 3:], axis=1)
            errors.append((lin_error, ang_error))
            print(f"action {i}, lin_error: {np.mean(lin_error):.3e} ang_error: {np.mean(ang_error):.3e}")
            i += 1

        fig, axs = plt.subplots(nrows=len(errors), ncols=2, figsize=(8, 8), dpi=150)
        for orbit_id in range(len(errors)):
            lin_err, ang_err = errors[orbit_id]
            ax_lin, ax_ang = axs[orbit_id, :]
            sns.histplot(x=lin_err, bins=50, stat="probability", ax=ax_lin, kde=True)
            sns.histplot(x=ang_err, bins=50, stat="probability", ax=ax_ang, kde=True)
            ax_lin.set_title(f"Lin error g:{orbit_id}")
            ax_ang.set_title(f"Ang error g:{orbit_id}")
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    robot_name = "anymal_c"

    dataset = ProprioceptiveDataset(
        robot_name=robot_name,
        type="train",
        n_total_samples=100000,
        angular_momentum=True,
        augment=False,
        kinetic_energy=True,
    )



from pathlib import Path

import numpy as np
import pybullet
from escnn.group import Group, Representation
from pybullet_utils.bullet_client import BulletClient
from scipy.spatial.transform import Rotation

from morpho_symm.data.DynamicsRecording import DynamicsRecording
from morpho_symm.utils.algebra_utils import permutation_matrix
from morpho_symm.utils.rep_theory_utils import escnn_representation_form_mapping, group_rep_from_gens
from morpho_symm.utils.robot_utils import load_symmetric_system


def get_kinematic_three_rep(G: Group):
    #  [0   1    2   3]
    #  [RF, LF, RH, LH]
    rep_kin_three = {G.identity: np.eye(4, dtype=int)}
    gens = [permutation_matrix([1, 0, 3, 2]), permutation_matrix([2, 3, 0, 1]), permutation_matrix([0, 1, 2, 3])]
    for h, rep_h in zip(G.generators, gens):
        rep_kin_three[h] = rep_h

    rep_kin_three = group_rep_from_gens(G, rep_kin_three)
    rep_kin_three.name = "kin_three"
    return rep_kin_three


def get_Rd_signals_on_kin_subchains(G: Group, rep_kin_three: Representation):
    rep_R3 = G.representations["R3"]
    rep_F = {G.identity: np.eye(12, dtype=int)}
    gens = [np.kron(rep_kin_three(g), rep_R3(g)) for g in G.generators]
    for h, rep_h in zip(G.generators, gens):
        rep_F[h] = rep_h

    rep_F = group_rep_from_gens(G, rep_F)
    rep_F.name = "R3_on_legs"
    return rep_F


def convert_mini_cheetah_raysim_recordings(data_path: Path):
    """Convertion script for the recordings of observations from the Mini-Cheetah Robot.

    This function takes recordings stored into a single numpy array of shape (time, state_dim) where the state is
    defined as [state]:
        base position,                           (3,)
        base_velocity,                           (3,)
        base_orientation,                        (3,)
        base_angular_velocity,                   (3,)
        feet_positions,                          (12,)
        joint position,                          (12,)
        joint velocities,                        (12,)
        joint torques,                           (12,)
        gait,                                    (4)
        reference base height                    (1,)
        reference base velocities,               (3,)
        reference base orientation,              (3,)
        reference base angular vel,              (3,)
        reference feet positions                 (12,)
        _________________________________________________________
        TOTAL:                                    85. Dimensions
    The conversion process takes these measurements and does the following:
        1. Stores them into a DynamicsRecording format, for easy loading
        2. Defines the group representation for each observation.
        3. Changes joint position to Pinocchio convention, used by MorphoSymm.
    """
    assert data_path.exists(), f"Path {data_path.absolute()} does not exist"
    state = np.load(data_path)
    assert state.shape[-1] == 86, f"Expected {86} dimensions in the state, got {state.shape[-1]}"
    # Load the Mini-Cheetah robot
    robot, G = load_symmetric_system(robot_name="mini_cheetah")
    rep_Q_js = G.representations["Q_js"]  # Representation on joint space position coordinates
    rep_TqQ_js = G.representations["TqQ_js"]  # Representation on joint space velocity coordinates
    rep_Rd = G.representations["R3"]  # Representation on vectors in R^d
    rep_Rd_pseudo = G.representations["R3_pseudo"]  # Representation on pseudo vectors in R^d
    rep_euler_xyz = G.representations["euler_xyz"]  # Representation on Euler angles
    rep_kin_three = get_kinematic_three_rep(G)  # Permutation of legs
    rep_Rd_on_limbs = get_Rd_signals_on_kin_subchains(G, rep_kin_three)  # Representation on R^3 on legs

    rep_z = group_rep_from_gens(G, rep_H={h: rep_Rd(h)[2, 2].reshape((1, 1)) for h in G.elements if h != G.identity})
    rep_z.name = "base_z"

    # Define observation variables and their group representations

    # Base body observations ___________________________________________________________________________________________
    base_pos = state[:, :3]  # Rep: rep_Rd
    base_z = state[:, [2]]  # Rep: rep_z
    base_vel = state[:, 3:6]  # Rep: rep_Rd
    base_ori = state[:, 6:9]  # Rep: rep_euler_xyz
    base_ang_vel = state[:, 9:12]  # Rep: rep_euler_xyz
    ref_base_z = state[:, [64]]  # Rep: rep_z
    ref_base_vel = state[:, 65:68]  # Rep: rep_Rd
    ref_base_ori = state[:, 68:71]  # Rep: rep_euler_xyz
    ref_base_ang_vel = state[:, 71:74]  # Rep: rep_euler_xyz
    base_z_error = base_z - ref_base_z  # Rep: rep_z
    base_vel_error = base_vel - ref_base_vel  # Rep: rep_Rd
    base_ang_vel_error = base_ang_vel - ref_base_ang_vel  # Rep: rep_euler_xyz
    base_ori_error = base_ori - ref_base_ori  # Rep: rep_euler_xyz
    # Define the representation of the rotation matrix R that transforms the base orientation.
    rep_rot_flat = {}
    # R = Rotation.from_euler("xyz", base_ori[2]).as_matrix()
    for h in G.elements:
        rep_rot_flat[h] = np.kron(rep_Rd(h), rep_Rd(~h).T)
    rep_rot_flat = escnn_representation_form_mapping(G, rep_rot_flat)
    rep_rot_flat.name = "SO(3)_flat"
    base_ori_R = np.asarray([Rotation.from_euler("xyz", ori).as_matrix() for ori in base_ori])
    base_ori_R_flat = base_ori_R.reshape(base_ori.shape[0], -1)

    # Euclidean space observations _____________________________________________________________________________________
    feet_pos = state[:, 12:24]  # Rep: rep_Rd_on_limbs
    gait = state[:, 60:64]  # Rep: rep_kin_three
    ref_feet_pos = state[:, 74:86]  # Rep: rep_Rd_on_limbs
    feet_pos_error = feet_pos - ref_feet_pos  # Rep: rep_Rd_on_limbs
    # Joint-Space observations _________________________________________________________________________________________
    joint_vel = state[:, 36:48]
    joint_torques = state[:, 48:60]
    # Joint positions need to be converted to the unit circle parametrization [cos(q), sin(q)].
    # For God’s sake, we need to avoid using PyBullet.
    bullet_client = BulletClient(connection_mode=pybullet.DIRECT)
    robot.configure_bullet_simulation(bullet_client=bullet_client)
    # Get zero reference position.
    q0, _ = robot.pin2sim(robot._q0, np.zeros(robot.nv))
    q_js_ms = state[:, 24:36] + q0[7:]  # Add offset to the measurements from UMich
    cos_q_js, sin_q_js = np.cos(q_js_ms), np.sin(q_js_ms)  # convert from angle to unit circle parametrization
    # Define joint positions [q1, q2, ..., qn] -> [cos(q1), sin(q1), ..., cos(qn), sin(qn)] format.
    q_js_unit_circle_t = np.stack([cos_q_js, sin_q_js], axis=2)
    q_js_unit_circle_t = q_js_unit_circle_t.reshape(q_js_unit_circle_t.shape[0], -1)
    joint_pos_S1, joint_pos_rep = q_js_unit_circle_t, rep_Q_js  # Joints in angle not unit circle representation
    joint_pos = q_js_ms  # Joints in angle representation

    # Subsample the data by skippig by ignoring odd frames. ============================================================
    dt_subsample = 3
    base_pos = base_pos[::dt_subsample]
    base_z = base_z[::dt_subsample]
    base_vel = base_vel[::dt_subsample]
    base_ori = base_ori[::dt_subsample]
    base_ori_R_flat = base_ori_R_flat[::dt_subsample]
    base_ang_vel = base_ang_vel[::dt_subsample]
    feet_pos = feet_pos[::dt_subsample]
    joint_pos = joint_pos[::dt_subsample]
    joint_pos_S1 = joint_pos_S1[::dt_subsample]
    joint_vel = joint_vel[::dt_subsample]
    joint_torques = joint_torques[::dt_subsample]
    gait = gait[::dt_subsample]
    ref_base_z = ref_base_z[::dt_subsample]
    ref_base_vel = ref_base_vel[::dt_subsample]
    ref_base_ori = ref_base_ori[::dt_subsample]
    ref_base_ang_vel = ref_base_ang_vel[::dt_subsample]
    ref_feet_pos = ref_feet_pos[::dt_subsample]
    feet_pos_error = feet_pos_error[::dt_subsample]
    base_z_error = base_z_error[::dt_subsample]
    base_vel_error = base_vel_error[::dt_subsample]
    base_ang_vel_error = base_ang_vel_error[::dt_subsample]
    base_ori_error = base_ori_error[::dt_subsample]
    # Define the dataset.
    data_recording = DynamicsRecording(
        description=f"Mini Cheetah {data_path.parent.parent.stem}",
        info=dict(num_traj=1, trajectory_length=state.shape[0]),
        dynamics_parameters=dict(dt=0.001 * dt_subsample, group=dict(group_name=G.name, group_order=G.order())),
        recordings=dict(
            base_pos=base_pos[None, ...].astype(np.float32),
            base_z=base_pos[None, :, [2]].astype(np.float32),
            base_vel=base_vel[None, ...].astype(np.float32),
            base_ori=base_ori[None, ...].astype(np.float32),
            base_ori_R_flat=base_ori_R_flat[None, ...].astype(np.float32),
            base_ang_vel=base_ang_vel[None, ...].astype(np.float32),
            feet_pos=feet_pos[None, ...].astype(np.float32),
            feet_pos_error=feet_pos_error[None, ...].astype(np.float32),
            joint_pos=joint_pos[None, ...].astype(np.float32),
            joint_pos_S1=joint_pos_S1[None, ...].astype(np.float32),
            joint_vel=joint_vel[None, ...].astype(np.float32),
            joint_torques=joint_torques[None, ...].astype(np.float32),
            gait=gait[None, ...].astype(np.float32),
            base_z_error=base_z_error[None, ...].astype(np.float32),
            base_vel_error=base_vel_error[None, ...].astype(np.float32),
            base_ang_vel_error=base_ang_vel_error[None, ...].astype(np.float32),
            base_ori_error=base_ori_error[None, ...].astype(np.float32),
        ),
        state_obs=(
            "joint_pos",
            "joint_vel",
            "base_z_error",
            "base_ori",
            "base_ori_error",
            "base_vel_error",
            "base_ang_vel_error",
        ),
        action_obs=("joint_torques",),
        obs_representations=dict(
            joint_pos=rep_TqQ_js,  # Joint-Space observations
            joint_pos_S1=rep_Q_js,
            joint_vel=rep_TqQ_js,
            joint_torques=rep_TqQ_js,
            # Base body observations
            base_pos=rep_Rd,
            base_z=rep_z,
            base_z_error=rep_z,
            base_vel=rep_Rd,
            base_vel_error=rep_Rd,
            base_ori=rep_euler_xyz,
            base_ori_R_flat=rep_rot_flat,
            base_ang_vel=rep_euler_xyz,
            base_ang_vel_error=rep_euler_xyz,
            base_ori_error=rep_euler_xyz,
            # Euclidean space observations
            feet_pos=rep_Rd_on_limbs,
            feet_pos_error=rep_Rd_on_limbs,
            gait=rep_kin_three,
        ),
        # Ensure the angles in the unit circle are not disturbed by the normalization.
        obs_moments=dict(
            joint_pos_S1=(
                np.zeros(q_js_unit_circle_t.shape[-1]),
                np.ones(q_js_unit_circle_t.shape[-1]),
            ),
            base_ori_R_flat=(
                np.zeros(base_ori_R_flat.shape[-1]),
                np.ones(base_ori_R_flat.shape[-1]),
            ),
        ),
    )

    # Compute the mean and variance of all observations considering symmetry constraints.
    for obs_name in data_recording.recordings.keys():
        if obs_name in data_recording.obs_moments:
            continue
        data_recording.compute_obs_moments(obs_name=obs_name)

    file_name = f"n_trajs={data_recording.info['num_traj']}-frames={data_recording.info['trajectory_length']}.pkl"
    data_recording.save_to_file(data_path.parent.parent / file_name)
    print(f"Dynamics Recording saved to {data_path.parent.parent / file_name}")


if __name__ == "__main__":
    terrains = ["flat", "uneven_easy", "uneven_medium", "uneven_hard_squares"]
    modes = ["forward_minus_0_4", "forward_minus_0_4_yawrate_0_4", "forward_minus_0_4_yawrate_minus_0_4"]
    for terrain in terrains:
        for mode in modes:
            data_path = Path(f"raysim_recordings/{terrain}/{mode}/heightmap_logger/state_reduced_nmpc.npy")
            convert_mini_cheetah_raysim_recordings(data_path)



import logging
from abc import ABC, abstractmethod
from typing import Collection, List, Optional, Tuple, Union

import numpy as np
import scipy

try:
    import pinocchio as pin
    from pinocchio import JointModelFreeFlyer, RobotWrapper
except ImportError as e:
    raise ImportError(
        "Pinocchio is an optional dependency of MorphoSymm which is not installed.\n"
        "Please install it using: \n"
        "\t cd <morpho_symm_root_dir> \n"
        "\t pip install -e '.[pin]'"
    ) from e
from robot_descriptions.loaders.pinocchio import load_robot_description as pin_load_robot_description
from scipy.linalg import inv

from morpho_symm.utils.algebra_utils import quat_xyzw_to_SO3

log = logging.getLogger(__name__)

NameList = Collection[str]
Vector = Collection[float]
State = Tuple[np.ndarray, np.ndarray]


class PinSimWrapper(ABC):
    def __init__(
        self,
        robot_name: str,
        endeff_names: Optional[NameList] = None,
        fixed_base=False,
        reference_robot: Optional["PinSimWrapper"] = None,
        hip_height=1.0,
        init_q=None,
        q_zero=None,
    ):
        """Initializes the wrapper.

        Args:
            robot_name (str): Name of this robot instance
            endeff_names (List[str]): Names of the end-effectors.
            fixed_base (bool, optional): Determines if the robot base if fixed. Defaults to False.
            reference_robot (PinBulletWrapper, optional): Instance to copy the pinocchio model from. Defaults to None.
            hip_height (float): Height of the hip of the robot. Hip height on resting configuration.
            init_q (List[float]): Initial configuration of the robot of length nq.
            q_zero (List[float]): Zero configuration. This can be different from the Zero config defined in the URDF.
        """
        self.name = str.lower(robot_name)
        self.fixed_base = fixed_base
        self._endeff_names = endeff_names
        self.hip_height = hip_height

        # Default to URDF values
        self._qj_high_limit, self._qj_low_limit, self._dqj_limit = None, None, None
        # Initialize Pinocchio Robot.
        self.load_pinocchio_robot(reference_robot)

        self.n_js = self.nq - 7
        self.nf = len(self.endeff_names) if self.endeff_names is not None else -1

        self._q0 = pin.neutral(self.pinocchio_robot.model) if q_zero is None else np.array(q_zero)
        self._new_neutral = q_zero is not None
        self._diff_neutral_conf = np.zeros(self.nv)
        assert self._q0.size == self.nq, f"Expected |q_0|=3+4+nj={self.nq}, but received len(q0)={len(q_zero)}"

        self._init_q = np.concatenate((np.zeros(6), [1], np.zeros(self.n_js))) if init_q is None else np.array(init_q)
        assert len(self._init_q) == self.nq, (
            f"Expected |q_init|=3+4+nj={self.nq}, but received |q_init|={self._init_q.size}"
        )

        # Mappings between joint names pin and bullet ids, and pinocchio generalized q and dq coordinates
        self.pin_joint_space = {}
        self.joint_space_names = []
        for joint, joint_name in zip(self.pinocchio_robot.model.joints, self.pinocchio_robot.model.names):
            if joint.idx_q == -1:
                continue  # Ignore universe
            if joint.nq == 7:
                continue  # Ignore floating-base
            pin_joint_type = joint.shortname()

            self.joint_space_names.append(joint_name)
            low_lim = self.pinocchio_robot.model.lowerPositionLimit[joint.idx_q : joint.idx_q + joint.nq]
            high_lim = self.pinocchio_robot.model.upperPositionLimit[joint.idx_q : joint.idx_q + joint.nq]
            self.pin_joint_space[joint_name] = JointWrapper(
                type=pin_joint_type,
                nq=joint.nq,
                nv=joint.nv,
                idx_q=joint.idx_q,
                idx_v=joint.idx_v,
                pos_limit_low=low_lim,
                pos_limit_high=high_lim,
            )
            # Handle initial configurations.
            default_q0 = pin.neutral(self.pinocchio_robot.model)
            # Compute tangent vector between the default and new neutral configuration.
            diff_neutral_conf = pin.difference(self.pinocchio_robot.model, default_q0, self._q0)
            self._new_neutral = not np.allclose(diff_neutral_conf[6:], 0)
            self._diff_neutral_conf = diff_neutral_conf

        self.joint_space_names = sorted(self.joint_space_names, key=lambda x: self.pin_joint_space[x].idx_q)
        self.joint_space = {}

    def get_state(self) -> State:
        """Fetch state of the system from a physics simulator and return pinocchio convention (q, dq).

        Obtains the system state from the simulator and returns the generalized positions and velocities,
        following the pinocchio convention for q and dq.

        Returns:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
            dq (ndarray): Generalized velocity coordinates of shape (self.nv,) in pinocchio convention.
        """
        q_sim, dq_sim = self.get_state_sim()
        q_pin, dq_pin = self.sim2pin(q_sim, dq_sim)
        q_pin_centered = self.center_state(q_pin)
        return q_pin_centered, dq_pin

    def get_joint_space_state(self, q: Optional[np.ndarray] = None, v: Optional[np.ndarray] = None) -> State:
        """Fetch joint-space state of the system from a physics simulator and return pinocchio convention (q, dq).

        Obtains the system joint-space state from the simulator and returns the generalized positions and velocities.

        TODO: This function should handle cases where the Euclidean space `Ed` in which the system evolvs in is not
          3-dimensional. Such as fixed based-robots or planar robots.

        Args:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
            v (ndarray): Generalized velocity coordinates of shape (self.nv,) in pinocchio convention.

        Returns:
            q_js (ndarray): Joint-Space generalized position coordinates of shape (|Q_J|) in pinocchio convention.
            v_js (ndarray): Joint-Space generalized velocity coordinates of shape (|TqQ_J) in pinocchio convention.
        """
        if q is None or v is None:
            q_sim, v_sim = self.get_state_sim()
            q_pin, v_pin = self.sim2pin(q_sim, v_sim)
            q_pin_centered = self.center_state(q_pin)
        else:
            q_pin_centered, v_pin = q, v
        return q_pin_centered[7:], v_pin[6:]  # TODO: handle non-3D cases

    @abstractmethod
    def get_state_sim(self) -> State:
        """Get the system state in simulator convention.

        Returns:
             q (ndarray): Generalized position coordinates of shape in simulator convention.
            dq (ndarray): Generalized velocity coordinates of shape in simulator convention.
        """

    def get_base_configuration(self) -> np.ndarray:
        """Return an SE(3) homogenous transformation matrix describing the orientation and position of the robot base.

        Returns:
            base_config (np.ndarray): (4x4) homogenous transformation matrix.
        """
        q, dq = self.get_state()
        rB_w = q[:3]  # Base Position in World Frame
        RB_w = quat_xyzw_to_SO3(q[3:7])  # Base Rotation in World Frame
        XB_w = np.eye(4)
        XB_w[:3, :3] = RB_w
        XB_w[:3, 3] = rB_w
        return XB_w

    def reset_state(self, q, v, update_pin=False) -> None:
        """Sets the state of the system in the simulator and Pinocchio if requested.

        Args:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
            v (ndarray): Generalized velocity coordinates of shape (self.nv,) in pinocchio convention.
            update_pin (bool): If True, update the pinocchio model with the new state.
        """
        q_centered = self.center_state(q)

        # q_2 = self.uncenter_state(q_centered)
        # qjs = np.rad2deg(q[7:])
        # qjs_c = np.rad2deg(q_centered[7:])
        # qjs_2 = np.rad2deg(q_2[7:])
        # assert np.allclose(q, q_2), f"{np.rad2deg(q - q_2)[7:]}"

        q_sim, dq_sim = self.pin2sim(q_centered, v)
        self.reset_state_sim(q_sim, dq_sim)
        if update_pin:
            self.update_pinocchio(q_centered, v)

    @abstractmethod
    def reset_state_sim(self, q, v) -> None:
        """Reset the state of the robot in simulation.

        Returns:
             q (ndarray): Generalized position coordinates of shape in simulator convention.
            dq (ndarray): Generalized velocity coordinates of shape in simulator convention.
        """

    @abstractmethod
    def sim2pin(self, q, dq) -> State:
        """Converts system state configuration from simulator to pinocchio convention.

        Args:
             q (ndarray): Generalized position coordinates of shape in simulator convention.
            dq (ndarray): Generalized velocity coordinates of shape in simulator convention.

        Returns:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
            dq (ndarray): Generalized velocity coordinates of shape (self.nv,) in pinocchio convention.
        """

    @abstractmethod
    def pin2sim(self, q, dq) -> State:
        """Converts system state configuration from simulator to pinocchio convention.

        Args:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
            dq (ndarray): Generalized velocity coordinates of shape (self.nv,) in pinocchio convention.

        Returns:
             q (ndarray): Generalized position coordinates of shape in simulator convention.
            dq (ndarray): Generalized velocity coordinates of shape in simulator convention.
        """

    def center_state(self, q) -> Vector:
        """If there is a defined Zero configuration this is used for centering the pinocchio state position coordinates.

        This function handles the centering of the state configuration position coordinates, considering the topology
        of the joints configuration space. That is if the joint is a continuous revolute or a spherical, the
        configuration of the joint is a point in the unit circle or a unit quaterion, respectively. In this case,
        the substraction of two points in these topological spaces has to be handled accordingly.

        Args:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.

        Returns:
            q_centered (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
        """
        if not self._new_neutral:
            return q
        # Compute tangent vector between current configuration and the default neutral configuration
        q_diff = pin.difference(self.pinocchio_robot.model, self.pinocchio_robot.q0, q)
        # Configuration w.r.t to new neutral configuration is the addition of the tangent vectors
        q_new_diff = self._diff_neutral_conf - q_diff
        q_centered = pin.integrate(self.pinocchio_robot.model, self.pinocchio_robot.q0, q_new_diff)
        q_centered[:7] = q[:7]
        return q_centered

    def uncenter_state(self, q) -> Vector:
        """If there is a defined Zero configuration this is used for uncenter the pinocchio state position coordinates.

        See center_state for more details.

        Args:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.

        Returns:
            q_centered (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
        """
        if not self._new_neutral:
            return q
        q_diff = pin.difference(self.pinocchio_robot.model, self.pinocchio_robot.q0, q)
        q_new_diff = -q_diff + self._diff_neutral_conf
        q_uncentered = pin.integrate(self.pinocchio_robot.model, self.pinocchio_robot.q0, q_new_diff)
        q_uncentered[:7] = q[:7]
        return q_uncentered

    def update_pinocchio(self, q, dq):
        """Updates the pinocchio robot.

        This includes updating:
        - kinematics
        - joint and frame Jacobian
        - Centroidal momentum

        Args:
            q (ndarray): Generalized position coordinates of shape (self.nq,) in pinocchio convention.
            dq (ndarray): Generalized velocity coordinates of shape (self.nv,) in pinocchio convention.
        """
        self.pinocchio_robot.forwardKinematics(q, dq)
        self.pinocchio_robot.framesForwardKinematics(q)

        pin.ccrba(self.pinocchio_robot.model, self.pinocchio_robot.data, q, dq)
        pin.computeKineticEnergy(self.pinocchio_robot.model, self.pinocchio_robot.data)
        pin.computePotentialEnergy(self.pinocchio_robot.model, self.pinocchio_robot.data)
        inv(self.pinocchio_robot.data.Ig) @ self.pinocchio_robot.data.hg

    def load_pinocchio_robot(self, reference_robot: Optional["PinSimWrapper"] = None):
        """Function to load and configure the pinocchio instance of your robot.

        Load the pinocchio robot from the URDF file and expose it for user use in the class property `pinocchio_robot`.


        Args:
            reference_robot: Instance of your robot class. If provided, the pinocchio robot will be copied from the
            reference avoiding copying the robot `model`.


        Returns:
            RobotWrapper: Instance of your pinocchio robot.
        """
        # if reference_robot is not None:
        #     import sys
        #     pin_robot = copy.copy(reference_robot.pinocchio_robot)
        #     # pin_robot.data = copy.copy(reference_robot.pinocchio_robot.data)
        #     assert sys.getrefcount(pin_robot.data) <= 2
        # else:
        pin_robot = pin_load_robot_description(f"{self.name}_description", root_joint=JointModelFreeFlyer())
        log.debug(f"Loaded pinocchio robot model for {self.name}. The robot's joints are:")
        n_pin_joints = len(pin_robot.model.joints)
        for idx, joint, joint_name in zip(range(n_pin_joints), pin_robot.model.joints, pin_robot.model.names):
            pin_joint_type = joint.shortname()
            log.debug(
                "\t -{:<3} {:<20} joint_type: {:<20} nq:{:<3} nv:{:<3} idx_q:{:<3} idx_v:{:<3}".format(
                    idx, joint_name, pin_joint_type, joint.nq, joint.nv, joint.idx_q, joint.idx_v
                )
            )

        self._mass = float(np.sum([i.mass for i in pin_robot.model.inertias]))  # [kg]
        self._pinocchio_robot = pin_robot

    @property
    def pinocchio_robot(self) -> RobotWrapper:
        """Get a handle of the pinocchio robot instance.

        Returns:
            pin_robot (RobotWrapper): Pinocchio robot instance.
        """
        if self._pinocchio_robot is None:
            raise AttributeError("Pinocchio robot has not been loaded")
        return self._pinocchio_robot

    @property
    def mass(self) -> float:
        """Get the total mass of the robot.

        Returns:
            mass (float): Total mass of the robot in [kg] computed from the robot's URDF file.
        """
        return float(np.sum([i.mass for i in self.pinocchio_robot.model.inertias]))  # [kg]

    @property
    def velocity_limits(self, q=None, dq=None) -> np.ndarray:
        """Get the velocity limits of the robot.

        Args:
            q: Generalized positions (nq,).
            dq: Generalized velocities (nv,).

        Returns:
            vel_lims (ndarray): Velocity limits of shape (self.nv,).
        """
        if self.pinocchio_robot is None:
            raise AttributeError("Pinocchio robot has not been loaded")
        return self.pinocchio_robot.model.velocityLimit[6:]

    @property
    def joint_pos_limits(self, q=None, dq=None):
        """Get the joint position limits of the robot.

        Args:
            q: generalized positions (nq,).
            dq: generalized velocities (nv,).

        Returns:
            qj_low_limit (ndarray): Lower joint position limits of shape (self.nj,).
        """
        if self.pinocchio_robot is None:
            raise AttributeError("Pinocchio robot has not been loaded")
        return self.pinocchio_robot.model.lowerPositionLimit[7:], self.pinocchio_robot.model.upperPositionLimit[7:]

    @property
    def endeff_names(self) -> Collection[str]:
        """Returns the names of the end-effectors of the robot.

        Returns:
            Collection: List of end-effector names.
        """
        return self._endeff_names

    def get_init_config(self, random=False, angle_sweep=None, fix_base=False):
        """Get initial configuration of the robot.

        Args:
            random (bool): if True, randomize the initial configuration.
            angle_sweep (float): if not None, randomize the initial configuration within the given angle sweep.
            fix_base (bool): if True, do not randomize base orientation.

        Returns:
            q (ndarray): generalized positions .
            dq (ndarray): generalized velocities.
        """
        q = self._init_q
        v = np.zeros(self.nv)

        if random:
            pitch = np.random.uniform(low=-np.deg2rad(25), high=np.deg2rad(25)) if not fix_base else 0
            roll = np.random.uniform(low=-np.deg2rad(25), high=np.deg2rad(25)) if not fix_base else 0
            yaw = np.random.uniform(low=-np.deg2rad(25), high=np.deg2rad(25)) if not fix_base else 0

            base_ori = scipy.spatial.transform.Rotation.from_euler("xyz", np.array([roll, pitch, yaw])).as_quat()

            for joint_name, joint in self.pin_joint_space.items():
                idx_q, idx_v = joint.state_idx
                q_j, v_j = joint.random_configuration(max_range=angle_sweep if angle_sweep is not None else np.pi)
                q[idx_q] = joint.add_configuration(q[idx_q], q_j)
                v[idx_v] += v_j
            q[3:7] = base_ori

        return q, v

    @property
    def nq(self) -> int:
        """The dimension of the generalized positions coordinates of the robot."""
        return self.pinocchio_robot.nq

    @property
    def nv(self) -> int:
        """The dimension of the generalized velocity coordinates of the robot."""
        return self.pinocchio_robot.nv

    @classmethod
    def from_instance(other: "PinSimWrapper") -> "PinSimWrapper":
        """Creates another instance of this robot wrapper without duplicating the model or data from pinocchio robot.

        This is usefull when we want to spawn multiple instances of the same robot on the physics simulator.


        Args:
            other (PinBulletWrapper): The instance from which to get the model and data for the pinocchio model

        Returns:
            PinBulletWrapper: A new instance of this robot wrapper
        """
        raise NotImplementedError("This method is not implemented yet")

    def __repr__(self):
        """."""
        return f"{self.name}-nq:{self.nq}-nv:{self.nv}"


class JointWrapper:
    def __init__(
        self,
        type: Union[str, int],
        idx_q: int,
        idx_v: int,
        nq: int,
        nv: int,
        pos_limit_low: Union[List[float], float] = -np.inf,
        pos_limit_high: Union[List[float], float] = np.inf,
    ):
        self.type = type
        self.idx_q = idx_q
        self.idx_v = idx_v
        self.nq = nq
        self.nv = nv
        self.pos_limit_low = pos_limit_low
        self.pos_limit_high = pos_limit_high

    def random_configuration(self, max_range=None):
        """Random configuration for the joint considering dimensions of the position space and tangent space.

        Returns:
            q (ndarray): Random configuration of the joint.
            v (ndarray): Zero velocity.
        """
        if self.nq == 1:
            theta = np.random.uniform(self.pos_limit_low, self.pos_limit_high)
            theta = min(max_range, theta)
            theta = max(-max_range, theta)
            return theta, np.array([0])
        if self.nq == 2 and self.nv == 1:  # Unit circle
            max_range = np.pi if max_range is None else max_range
            theta = np.random.uniform(-max_range, max_range)
            return np.array([np.cos(theta), np.sin(theta)]), np.array([0])
        if self.nq == 4 and self.nv == 3:  # Spherical joint
            # Generate random quaternion
            u = np.random.uniform(-1, 1, size=3)
            theta1 = max_range * u[0]
            theta2 = max_range * u[1]
            theta3 = max_range * u[2]

            w = np.cos(theta3 / 2) * np.sin(theta2 / 2) * np.sin(theta1 / 2) + np.sin(theta3 / 2) * np.cos(
                theta2 / 2
            ) * np.cos(theta1 / 2)
            x = np.sin(theta3 / 2) * np.cos(theta2 / 2) * np.sin(theta1 / 2) - np.cos(theta3 / 2) * np.sin(
                theta2 / 2
            ) * np.cos(theta1 / 2)
            y = np.cos(theta3 / 2) * np.cos(theta2 / 2) * np.sin(theta1 / 2) + np.sin(theta3 / 2) * np.sin(
                theta2 / 2
            ) * np.cos(theta1 / 2)
            z = np.cos(theta3 / 2) * np.sin(theta2 / 2) * np.cos(theta1 / 2) - np.sin(theta3 / 2) * np.cos(
                theta2 / 2
            ) * np.sin(theta1 / 2)

            return np.array([x, y, z, w]), np.array([0, 0, 0])

    def zero(self) -> State:
        if self.nq == 1:
            return np.array([0]), np.array([0])
        if self.nq == 2 and self.nv == 1:  # Unit circle
            return np.array([1, 0]), np.array([0])
        if self.nq == 4 and self.nv == 3:  # Spherical joint
            return np.array([0, 0, 0, 1]), np.array([0, 0, 0])

    def add_configuration(self, q1, q2) -> State:
        # TODO: handle these through pinocchio
        if self.nq == 1:
            return q1 + q2
        if self.nq == 2 and self.nv == 1:  # Unit circle
            theta1 = np.arctan2(q1[1], q1[0])
            theta2 = np.arctan2(q2[1], q2[0])
            theta = theta1 + theta2
            return np.array([np.cos(theta), np.sin(theta)])
        if self.nq == 4 and self.nv == 3:  # Spherical joint
            raise NotImplementedError()

    def substract_configuration(self, q1, q2) -> State:
        # TODO: handle these through pinocchio
        if self.nq == 1:
            return q1 - q2
        if self.nq == 2 and self.nv == 1:  # Unit circle
            theta1 = np.arctan2(q1[1], q1[0])
            theta2 = np.arctan2(q2[1], q2[0])
            theta = theta1 - theta2
            return np.array([np.cos(theta), np.sin(theta)])
        if self.nq == 4 and self.nv == 3:  # Spherical joint
            raise NotImplementedError()

    @property
    def state_idx(self) -> Tuple[List[int], List[int]]:
        idx_q = list(range(self.idx_q, self.idx_q + self.nq))
        idx_v = list(range(self.idx_v, self.idx_v + self.nv))
        return idx_q, idx_v

    def __repr__(self):
        return f"[{self.type}]-nq:{self.nq}-nv:{self.nv}-idx_q:{self.idx_q}-idx_v:{self.idx_v}"


class SimPinJointWrapper(ABC):
    def __init__(self) -> object:
        self.pin_joint = None
        self.sim_joint = None

    def sim2pin(self, q, dq) -> State:
        return q, dq

    def pin2sim(self, q, dq) -> State:
        return q, dq

    def __repr__(self):
        return (
            f"Pin[{self.pin_joint.type}]_"
            f"idxq:{list(range(self.pin_joint.idx_q, self.pin_joint.idx_q + self.pin_joint.nq))}_"
            f"idxv:{list(range(self.pin_joint.idx_v, self.pin_joint.idx_v + self.pin_joint.nv))}-"
            f"Sim[{self.sim_joint.type}]_"
            f"idxq:{list(range(self.sim_joint.idx_q, self.sim_joint.idx_q + self.sim_joint.nq))}_"
            f"idxv:{list(range(self.sim_joint.idx_v, self.sim_joint.idx_v + self.sim_joint.nv))}"
        )



import logging
import time
from pathlib import Path

import hydra
import numpy as np
import scipy
from escnn.group import Group, Representation
from omegaconf import DictConfig
from pynput import keyboard
from scipy.spatial.transform import Rotation
from tqdm import tqdm
from utils.pybullet_visual_utils import draw_plane, draw_vector, render_camera_trajectory, spawn_robot_instances
from utils.robot_utils import load_symmetric_system

import morpho_symm
from morpho_symm.data.DynamicsRecording import DynamicsRecording
from morpho_symm.utils.algebra_utils import matrix_to_quat_xyzw, permutation_matrix
from morpho_symm.utils.pybullet_visual_utils import configure_bullet_simulation
from morpho_symm.utils.rep_theory_utils import group_rep_from_gens

log = logging.getLogger(__name__)


def on_key_press(key):
    """TODO: Remove from here."""
    try:
        # Get the pressed key and add it to the global list of new characters
        if key.char.isdigit():
            num_pressed.append(int(key.char))
        else:
            new_command.append(key.char)
    except AttributeError:
        # Ignore special keys without a char attribute
        pass


# Set up the listener and the list to store new characters
new_command = []
num_pressed = []
listener = keyboard.Listener(on_press=on_key_press)
listener.start()


def get_kinematic_three_rep(G: Group):
    #  [0   1    2   3]
    #  [RF, LF, RH, LH]
    rep_kin_three = {G.identity: np.eye(4, dtype=int)}
    gens = [permutation_matrix([1, 0, 3, 2]), permutation_matrix([2, 3, 0, 1]), permutation_matrix([0, 1, 2, 3])]
    for h, rep_h in zip(G.generators, gens):
        rep_kin_three[h] = rep_h

    rep_kin_three = group_rep_from_gens(G, rep_kin_three)

    return rep_kin_three


def get_ground_reaction_forces_rep(G: Group, rep_kin_three: Representation):
    rep_R3 = G.representations["R3"]
    rep_F = {G.identity: np.eye(12, dtype=int)}
    gens = [np.kron(rep_kin_three(g), rep_R3(g)) for g in G.generators]
    for h, rep_h in zip(G.generators, gens):
        rep_F[h] = rep_h

    rep_F = group_rep_from_gens(G, rep_F)
    return rep_F


def load_mini_cheetah_trajs(data_path: Path):
    """Load the mini-cheetah trajectories from the .mat files.

    The Umich dataset contains in the following order [RF, LF, RH, LH]:
    - q: joint encoder value (num0_data,12)
    - qd: joint angular velocity (num_data,12)
    - p: foot position from FK (num_data,12)
    - v: foot velocity from FK (num_data,12).

    - imu_acc: linear acceleration from imu (num_data,3)
    - imu_omega: angular velocity from imu (num_data,3)
    - contacts: contact data (num_data,4)
    - tau_est (optional): estimated control torque (num_data,12)
    - F (optional): ground reaction force
    Args:
        data_path (Path): Path to the .mat files.

    Returns:
        recordings (dict): Dictionary of recording name and data.
    """
    assert data_path.exists(), f"Path {data_path} does not exist."
    recordings = {}
    for data_name in data_path.glob("*.mat"):
        raw_data = scipy.io.loadmat(data_name)
        recordings[data_name.stem] = raw_data
    return recordings


def update_contacts(pb, feet_pos, prev_contact_state, contact_state, planes=None):
    out_of_view_pos = [0, 0, 100]

    if planes is None:
        planes = []
        for r in feet_pos:
            # Draw a dark purple plane (R, r) in the world frame.
            planes.append(
                draw_plane(pb, np.eye(3), out_of_view_pos, color=(0.9, 0, 0.9, 0.5), cylinder=True, size=(0.05, 0.005))
            )

    state_change = contact_state != prev_contact_state
    idx = np.where(state_change)[0]
    for i in idx:
        plane_pos = feet_pos[i] if contact_state[i] else out_of_view_pos
        pb.resetBasePositionAndOrientation(planes[i], plane_pos, np.array([0, 0, 0, 1]))

    stable_contact = contact_state == prev_contact_state
    idx = np.where(np.logical_and(stable_contact, contact_state))[0]
    for i in idx:
        plane_pos = feet_pos[i]
        pb.resetBasePositionAndOrientation(planes[i], plane_pos, np.array([0, 0, 0, 1]))

    return planes


def update_heading(pb, X_B, heading_arrow=None):
    pos = (X_B @ np.array([-0.09, 0, 0.045, 1]))[:3]
    if heading_arrow is None:
        heading_arrow = draw_plane(pb, X_B[:3, :3], pos, color=(0.504, 0.931, 0.970, 1.0), size=(0.01, 0.04, 0.01))
    else:
        pb.resetBasePositionAndOrientation(heading_arrow, pos, matrix_to_quat_xyzw(X_B[:3, :3]))
    return heading_arrow


def update_ground_reaction_forces(pb, feet_pos, forces, contact_state, vectors=None):
    vectors = [None] * 4 if vectors is None else vectors

    feet_in_contact = np.where(np.isclose(contact_state, 1))[0]
    for idx in feet_in_contact:
        # Remove old vector if present
        if vectors[idx] is not None:
            pb.removeBody(vectors[idx])
        # Draw a light purple vector representing the force in world frame.
        r_w, f_w = feet_pos[idx], forces[idx]
        force_vect = draw_vector(pb, origin=r_w, vector=f_w, v_color=(0.5, 0, 0.5, 0.5), scale=0.005)
        vectors[idx] = force_vect

    # Remove forces of legs not in contact
    feet_not_in_contact = np.where(np.isclose(contact_state, 0))[0]
    for idx in feet_not_in_contact:
        if vectors[idx] is not None:
            pb.removeBody(vectors[idx])
            vectors[idx] = None

    return vectors


@hydra.main(config_path="cfg", config_name="config_visualization", version_base="1.3")
def main(cfg: DictConfig):
    """Visualize the effect of DMSs transformations in 3D animation.

    This script visualizes the DMSs transformations on robot state and on proprioceptive and exteroceptive measurements.
    """
    cfg.robot.seed = cfg.robot.seed if cfg.robot.seed >= 0 else np.random.randint(0, 1000)
    np.random.seed(cfg.robot.seed)

    if "mini_cheetah" not in cfg.robot.name:
        raise NotImplementedError("For the moment we have only real-world data from Mini-Cheetah.")

    # Get robot instance, along with representations of the symmetry group on the Euclidean space (in which the robot
    # base B evolves in) and Joint Space (in which the internal configuration of the robot evolves in).
    robot, G = load_symmetric_system(robot_cfg=cfg.robot, debug=cfg.debug)

    rep_Q_js = G.representations["Q_js"]
    # rep_TqJ = G.representations['TqQ_js']
    rep_E3 = G.representations["Ed"]
    rep_R3 = G.representations["R3"]

    offset = max(0.2, 1.8 * robot.hip_height)
    base_pos = np.array([-offset if G.order() != 2 else 0, -offset] + [robot.hip_height * 5.5])
    pb = configure_bullet_simulation(gui=cfg.gui, debug=cfg.debug)
    # Define the positions to spawn the |G| instances of the robot to represent the |G| symmetric states.
    X_B = np.eye(4)
    X_B[:3, 3] = base_pos
    orbit_X_B = [rep_E3(g) @ X_B @ np.linalg.inv(rep_E3(g)) for g in G.elements]
    robots = spawn_robot_instances(
        robot,
        bullet_client=pb,
        base_positions=[X[:3, 3] for X in orbit_X_B],
        tint=cfg.robot.tint_bodies,
    )
    robot = robots[0]
    end_effectors = robot.bullet_ids_allowed_floor_contacts

    # Load a trajectory of motion and measurements from the mini-cheetah robot
    recordings_path = (
        Path(morpho_symm.__file__).parent
        / "data/mini_cheetah/raysim_recordings/flat/forward_minus_0_4/n_trajs=1-frames=7771-train.pkl"
    )
    dyn_recordings = DynamicsRecording.load_from_file(recordings_path)
    # Load and prepare data for visualization
    q_js_t = dyn_recordings.recordings["joint_pos"]  # Joint space positions
    v_js_t = dyn_recordings.recordings["joint_vel"]  # Joint space velocities
    base_ori_t = dyn_recordings.recordings["base_ori"][0]  # Base orientation
    feet_pos = dyn_recordings.recordings["feet_pos"]  # Feet positions  [x,y,z] w.r.t base frame
    # ground_reaction_forces = recpitcording['F']
    # feet_contact_states = recording['contacts']
    # Prepare representations acting on proprioceptive and exteroceptive measurements.
    rep_kin_three = dyn_recordings.obs_representations["gait"]  # Contact state is a 4D vector.
    rep_grf = dyn_recordings.obs_representations["ref_feet_pos"]  #

    q0, _ = robot.pin2sim(robot._q0, np.zeros(robot.nv))

    dt = dyn_recordings.dynamics_parameters["dt"]
    timestamps = np.linspace(0, dt * q_js_t.shape[0], q_js_t.shape[0])
    frames = []
    robot_frames = [[] for _ in range(G.order())]
    symmetry_frames = []
    robot_terrain, robot_grf_vects = [None for _ in range(G.order())], [None for _ in range(G.order())]
    robot_heading = [None for _ in range(G.order())]
    prev_contact_state = np.asarray([0, 0, 0, 0], dtype=np.uint8)

    fps = 30
    prev_t = timestamps[0]
    t0 = timestamps[0]
    run_time0 = time.time()

    camera_view_point = Rotation.from_euler("xyz", [0, -25, 90], True)
    # rpy = camera_view_point.as_euler("xyz") * 180/np.pi
    for i, t in tqdm(enumerate(timestamps), total=len(timestamps), desc=f"Capturing frames from {cfg.robot.name}"):
        # if i < 21000 or i > 25000:
        #     continue
        # for i, t in tqdm(enumerate(timestamps[:int(len(timestamps) // 2)]), total = len(timestamps), desc=f"Capturing frames from {cfg.robot.name}"):
        run_time = time.time()
        if cfg.gui:
            if run_time - run_time0 < t - t0:
                time.sleep(((t - t0) - (run_time - run_time0)) / 2)
                continue
        else:
            if t - prev_t < (1 / fps):
                continue
        prev_t = t

        # Define the recording base configuration.
        X_B[:3, :3] = Rotation.from_euler("xyz", base_ori_t[i] - np.array([0, 0, 0 * np.pi / 2])).as_matrix()
        X_B[:3, 3] = base_pos
        # contact_state = feet_contact_states[i]
        orbit_X_B, orbit_cam_target = {}, {}
        q_rec = np.concatenate([X_B[:3, 3], matrix_to_quat_xyzw(X_B[:3, :3]), q_js_t[i]])
        v_rec = np.concatenate([np.zeros(6), v_js_t[i]])
        # Just for Mit Cheetah.
        q = q_rec
        feet_pos_fk = feet_pos[i]
        a = feet_pos_fk[:3]
        for robot_idx, g in enumerate(G.elements):
            g_X_B = np.real(rep_E3(g) @ X_B @ rep_E3(~g))
            orbit_X_B[g] = g_X_B
            g_q_js = np.real(rep_Q_js(g) @ q[7:])
            g_q = np.concatenate([g_X_B[:3, 3], matrix_to_quat_xyzw(g_X_B[:3, :3]), g_q_js]).astype(float)

            # Set the robot configuration to symmetric state.
            robots[robot_idx].reset_state(g_q, np.zeros(robot.nv))

            # Get the position of the 4 feet of the robot in the world frame [RF, LF, RH, LH].
            # g_rf_w = [pb.getLinkState(robots[robot_idx].robot_id, name)[0] for name in end_effectors]

            # Display contact terrain
            # g_contact_state = np.rint(np.real(rep_kin_three(g) @ contact_state)).astype(np.uint8)
            # g_prev_contact_state = np.rint(np.real(rep_kin_three(g) @ prev_contact_state)).astype(np.uint8)(np.uint8)
            # robot_terrain[robot_idx] = update_contacts(pb,
            #                                            g_rf_w,
            #                                            g_prev_contact_state,
            #                                            g_contact_state,
            #                                            planes=robot_terrain[robot_idx])
            robot_heading[robot_idx] = update_heading(pb, g_X_B, heading_arrow=robot_heading[robot_idx])
            # Display ground reaction forces
            if not cfg.gui:
                # If in contacts draw the vect representing the ground reaction forces.
                # grf = ground_reaction_forces[i]
                # g_grf = np.real((rep_F(g) @ grf))
                # g_forces = np.split(g_grf, 4)

                # robot_grf_vects[robot_idx] = update_ground_reaction_forces(pb,
                #                                                            g_rf_w,
                #                                                            g_forces,
                #                                                            g_contact_state,
                #                                                            vectors=robot_grf_vects[robot_idx])

                g_camera_view_point = np.real(rep_R3(g) @ camera_view_point.as_matrix() @ np.linalg.inv(rep_R3(g)))
                roll, pitch, yaw = Rotation.from_matrix(g_camera_view_point).as_euler("xyz") * 180 / np.pi
                cam_target = orbit_X_B[g][:3, 3]
                robot_frames[robot_idx].append(
                    render_camera_trajectory(
                        pb,
                        roll=[roll],
                        pitch=[-10 if robot_idx in [0, 1, 4, 5] else 20],
                        yaw=[yaw + 90],
                        render_width=512,
                        render_height=512,
                        n_frames=1,
                        cam_distance=robot.hip_height * 3.0,
                        farPlane=robot.hip_height * 5.0,
                        cam_target_pose=cam_target,
                    )
                )

        if not cfg.gui:
            symmetry_frame = np.concatenate([robot_frames[robot_idx][-1] for robot_idx in range(G.order())], axis=2)
            symmetry_frames.append(symmetry_frame.squeeze())

            # frames.extend(render_camera_trajectory(pb, roll=[0], pitch=[-25], yaw=[0], n_frames=1,
            #                                        render_width=1000, render_height=500,
            #                                        cam_distance=robot.hip_height * 5.0,
            #                                        cam_distance=robot.hip_height * 3.5,
            #                                        cam_target_pose=[0, 0, robot.hip_height]))
        # prev_contact_state = contact_state

    frames = symmetry_frames

    if len(frames) > 0:
        from moviepy.editor import ImageSequenceClip

        # Save animation
        root_path = Path(morpho_symm.__file__).parents[1].absolute()
        save_path = root_path / "paper/tests"
        file_name = f"mini-cheetah_{G}-{recording_name}_dynamic_symmetries"
        file_path = save_path / f"{file_name}.gif"
        file_count = 1
        while file_path.exists():
            file_path = save_path / f"{file_name}({file_count}).gif"
            file_count += 1
        clip = ImageSequenceClip(list(frames), fps=fps // 2)
        clip.write_gif(file_path, fps=fps // 2, loop=0, fuzz=0.9)
        print(f"Animation with {len(frames)} ({len(frames) / fps // 2}[s]) saved to {file_path.absolute()}")

    pb.disconnect()


if __name__ == "__main__":
    main()
